from enum import Enum


class LambdaStates(Enum):
    TIMED_OUT = "Lambda timed out."
    SUCCEEDED = "Lambda process succeeded."
    FAILED = "Lambda process failed."


class EventTypes(Enum):
    STREAM = "stream_event"
    TASK = "task_event"
    SCHEDULER = "scheduler_event"
    GENERIC = "generic_event"


class Environment(Enum):
    QA = "qa"
    STAGING = "staging"
    PRODUCTION = "production"
    LOCAL = "local"



from worker.data.operations import get_data_by_path, nanround


class Element:
    """
    This is a base object for any element in the wellbore:
    - Pipe, and Bit
    - Hole section
    - Annulus section
    """

    def __init__(self, **kwargs):
        """
        Set the element properties through kwargs. The keys are:
        - 'top_depth'
        - 'length'
        - 'bottom_depth'
        """
        self.top_depth = None
        self.bottom_depth = None
        self.length = None

        if {'top_depth', 'length'} <= kwargs.keys():
            self.top_depth = get_data_by_path(kwargs, 'top_depth', func=float)
            self.length = get_data_by_path(kwargs, 'length', func=float)
            self.set_bottom_depth()
        elif {'top_depth', 'bottom_depth'} <= kwargs.keys():
            self.top_depth = get_data_by_path(kwargs, 'top_depth', func=float)
            self.bottom_depth = get_data_by_path(kwargs, 'bottom_depth', func=float)
            self.set_length()
        elif 'length' in kwargs:
            self.length = get_data_by_path(kwargs, 'length', func=float)
        else:
            raise KeyError("Missing keys: 'top_depth', 'bottom_depth', and 'length'")

    def set_length(self, length: float = None):
        if length is None:
            length = self.bottom_depth - self.top_depth
        self.length = length

    def set_bottom_depth(self, bottom_depth: float = None):
        if bottom_depth is None:
            bottom_depth = self.top_depth + self.length
        self.bottom_depth = bottom_depth

    def set_top_depth(self, top_depth: float = None):
        if top_depth is None:
            top_depth = self.bottom_depth - self.length
        self.top_depth = top_depth

    @property
    def area(self):
        return self.compute_get_area()

    def compute_get_area(self):
        """
        Compute the area of the element
        :return: area of the element in INCH^2
        :return:
        """
        raise NotImplementedError()

    @property
    def volume(self):
        return self.compute_get_volume()

    def compute_get_volume(self):
        """
        Compute the volume of the element
        :return: volume in FOOT^3
        """
        raise NotImplementedError()

    def cmp(self, measured_depth: float) -> int:
        """
        To check the position of the given measured depth vs the current section
        if measured depth is closer to surface compared to the annulus it returns -1,
        and if deeper it returns +1 and otherwise it is 0
        :param measured_depth:
        :return: -1: above, 0: at, +1: below
        """
        if measured_depth <= self.top_depth:
            return -1

        if self.bottom_depth < measured_depth:
            return +1

        return 0

    def __contains__(self, measured_depth: float):
        """
        To check if this element covers the given measured depth.
        :param measured_depth:
        :return:
        """
        return self.top_depth < measured_depth <= self.bottom_depth

    def __repr__(self):
        return f"{nanround(self.length, 2):>8} ft --> {nanround(self.top_depth, 2):>8} ft - {nanround(self.bottom_depth, 2):>8} ft"



"""
The purpose of this App is to test the StateMixin functionality
This app tests saving/loading, deleting and migration of states on two different storage types:
1. Mongo `temporary-state` collection
2. Redis
"""
from testcase.app.app_constants import constants  # noqa: F401



# flake8: noqa: E402

# added the main directory as the first path
import json
from os.path import dirname
import sys

parent = dirname(dirname(__file__))
sys.path.insert(0, parent)

from dotenv import load_dotenv

from testcase.app import app_lambda
from worker import API
from worker.test import AppTestRun
from worker.test.lambda_function_test_run import generate_runner_parser
from worker.test.local_testing.to_local_transfer import ToLocalTransfer, generate_transfer_parser


load_dotenv(override=True)


if __name__ == '__main__':
    # ========== INPUTS ========== #
    # Inputs for local data transfer app
    source_env = 'qa'
    source_asset_id = 37070404
    start_timestamp = 1502586000
    end_timestamp = 1502587000

    # you don't need to update the followings, the default ones are used.
    config_collections = []
    main_collections = []

    # INPUTS for runner
    timestep = 200
    to_delete = True
    storage_type = 'redis'

    collections = ['drilling-efficiency.mse']

    # ===== To Local Transfer Run ===== #
    parser = generate_transfer_parser()
    args = parser.parse_args([
        '--source_environment', source_env,
        '--source_asset_id', str(source_asset_id),
        '--config_collections', json.dumps(config_collections),
        '--main_collections', json.dumps(main_collections),
        '--start_timestamp', str(start_timestamp),
        '--end_timestamp', str(end_timestamp)
    ])
    mover = ToLocalTransfer(args)
    mover.run()

    # ===== Runner ===== #
    parser = generate_runner_parser()
    args = parser.parse_args([
        '--environment', 'local',
        '--asset_id', str(mover.local_asset_id),
        '--start_timestamp', str(start_timestamp),
        '--end_timestamp', str(end_timestamp),
        '--timestep', str(timestep),
        '--to_delete', str(to_delete),
        '--storage_type', storage_type
    ])

    def run():
        # loading of constants should happen here to avoid conflicts
        from testcase.app.app_constants import constants
        app = AppTestRun(app_lambda.lambda_handler, collections, constants, args=args)
        app.run()

    run()
    api = API()
    res = api.get(
        path="/v1/data/corva", collection='drilling-efficiency.mse',
        asset_id=mover.local_asset_id, sort="{timestamp: 1}", limit=1000
    )

    assert res.count > 10



import json
import os
from typing import List

from cached_property import cached_property
import redis

from worker.state.enums import StateStorageType
from worker.state.handlers.__init__ import StateHandler


class RedisStateHandler(StateHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.state_storage_type = StateStorageType.REDIS
        state_storage_limit_fatal = os.getenv("STATE_STORAGE_CRITICAL_LIMIT", 1_000.0)  # 1 MB
        state_storage_limit_warning = os.getenv("STATE_STORAGE_WARNING_LIMIT", 100.0)  # 100 kB
        self.set_size_limit("fatal", float(state_storage_limit_fatal))
        self.set_size_limit("warning", float(state_storage_limit_warning))

    @cached_property
    def redis_connection(self):
        """
        Setup redis and get connection
        :return:
        """
        cache_url = os.getenv("CACHE_URL", None)

        if not cache_url:
            raise Exception("redis key (CACHE_URL) not found in Environment Variables.")
        _redis = redis.Redis.from_url(cache_url)

        if not _redis:
            raise Exception("Could not connect to Redis with URL: {cache_url}")

        return _redis

    def load_state(self, asset_id: int, state_key: str) -> dict:
        """
        Load state from redis
        :return:
        """
        state = self.redis_connection.get(state_key)
        if state:
            return json.loads(state)

        return {}

    def save_state(self, asset_id: int, state: dict, state_key: str, timestamp: [int, None]):
        """
        Save the state to redis
        :param asset_id:
        :param state:
        :param state_key:
        :param timestamp:
        :return:
        """
        self.redis_connection.set(state_key, value=json.dumps(state))

    def migrate_state(self, asset_id: int, state_key: str, from_state_storage=StateStorageType.MONGO) -> dict:
        """
        Attempt to get state from mongo and save it to redis
        This is needed for assets that already running, and saving state to redis
        Useful if the lambda has been setup for mongo but previous live assets are saving to redis
        :param asset_id:
        :param state_key:
        :param from_state_storage:
        :return:
        """
        if from_state_storage != StateStorageType.MONGO:
            return {}

        from worker.state.handlers.mongo import MongoStateHandler
        state_handler = MongoStateHandler()

        state = state_handler.load_state(asset_id, state_key)
        if not state:
            state = {}

        self.save_state(asset_id, state, state_key, None)
        return state

    def delete_states(self, asset_id: int, state_keys: List[str]):
        """
        Delete the states corresponding to given keys
        :param asset_id:
        :param state_keys: a list of redis keys
        :return:
        """
        self.redis_connection.delete(*state_keys)



import json
import time

from worker.app.modules.time_activity_module import TimeActivityModule
from worker.data.operations import gather_data_for_period
from worker.framework import constants
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state.mixins import StateMixin
from worker.state.state import State
from worker import exceptions


class App(StateMixin, LoggingMixin, RollbarMixin):
    app_state_fields = {
        'asset_id': int,
        'last_processed_timestamp': int
    }

    def __init__(self, *args, **kwargs):
        self.app_key = constants.get('global.app-key')
        self.app_name = constants.get('global.app-name')
        self.module_key = None

        state_storage_type = constants.get("global.state_storage_type", None)
        if state_storage_type is not None:
            # Add to kwargs if defined in constants, else will get from environment
            kwargs["state_storage_type"] = state_storage_type

        self.asset_id = None
        self.event = None

        super().__init__(*args, **kwargs)

    def load(self, event: str):
        """
        :param event: a scheduler event or wits stream
        :return:
        """
        event_type = self.get_event_type()
        valid_stream_collections = self.get_valid_stream_collections()
        event = self.clean_event(event)

        max_lookback_seconds = self.get_max_lookback_seconds()
        event = self.format_event(event_type, event)
        self.asset_id = self.determine_asset_id(event)

        self.state = self.load_state()

        event = self.load_event(event_type, event, max_lookback_seconds)
        self.event = self.filter_event_for_collections(event_type, valid_stream_collections, event)
        self.log_event(self.event, max_lookback_seconds)

    def log_event(self, event, max_lookback_seconds):
        self.debug(self.asset_id, "WITS input to {0} -> {1}".format(self.app_name, event))

        if not event:
            return

        batch_size = len(event)
        start_time = event[0].get("timestamp")
        end_time = event[-1].get("timestamp")
        system_time = time.time()

        self.debug(
            self.asset_id,
            "Received {} elements from {} to {} at {}. {} seconds of initial data are lookback.".format(
                batch_size, start_time, end_time, system_time, max_lookback_seconds
            )
        )

    @staticmethod
    def get_event_type():
        event_type = constants.get('global.event-type', "")

        if event_type not in ("scheduler", "wits_stream"):
            raise Exception('event_type specified incorrectly: {}'.format(event_type))

        return event_type

    @staticmethod
    def get_valid_stream_collections():
        valid_collections = constants.get('global.valid-stream-collections', [])
        if not isinstance(valid_collections, (str, list)):
            raise TypeError("Incorrect type of valid-stream-collections in global constants")

        if isinstance(valid_collections, str):
            valid_collections = [valid_collections]
        return valid_collections

    @staticmethod
    def clean_event(event):
        if not event:
            raise Exception("Empty event")

        try:
            if isinstance(event, (str, bytes, bytearray)):
                event = json.loads(event)
        except ValueError:
            raise Exception("Invalid event JSON")

        return event

    def get_max_lookback_seconds(self):
        """
        For each module (mostly in time-base modules), the time of processing does not
        match the last time of the event so extra data is required to look back and get
        the data so the processing can start from where it left off.
        :return:
        """

        time_modules = [module for module in self.get_modules() if issubclass(module, TimeActivityModule)]
        maximum_lookback = 0
        for module in time_modules:
            module_lookback = constants.get('{}.{}.export-duration'.format(self.app_key, module.module_key), default=0)
            maximum_lookback = max(module_lookback, maximum_lookback)

        return maximum_lookback

    def load_event(self, event_type, event, max_lookback_seconds):
        if event_type == 'scheduler':
            return self.load_scheduler_event(self.asset_id, event, max_lookback_seconds)

        if event_type == 'wits_stream':
            return self.load_wits_stream_event(self.asset_id, event, max_lookback_seconds)

        return None

    @staticmethod
    def filter_event_for_collections(event_type: str, valid_stream_collections: list, event: list):
        """
        This function filters the incoming event based on a list of valid collections

        :param event_type: Type of incoming event.
        :param valid_stream_collections: List of valid collections
        :param event: List of data records
        :return: List of records whose collection is one of the allowed event collections
        """

        # If event type is scheduler or event_collections is empty, return the entire event
        if event_type == 'scheduler' or not valid_stream_collections:
            return event

        # Filtering the event based on valid collections
        event = [record for record in event if record.get("collection") in valid_stream_collections]
        return event

    def load_scheduler_event(self, asset_id, event, max_lookback_seconds):
        """
        To load a scheduler event and get the wits stream data
        :param asset_id: The asset to load
        :param event: The cleaned event from Kafka scheduler stream
        :param max_lookback_seconds: Maximum amount of time to look back prior to the scheduler event to cover gaps
        :return: list of WITS data between the last processed timestamp and the final event item timestamp
        """

        start_timestamp = self.state.get('last_processed_timestamp', event[0].get('timestamp') - 1)
        end_timestamp = event[-1].get('timestamp')

        # the event is converted from scheduler to wits stream
        scheduler_event = gather_data_for_period(
            asset_id=asset_id,
            start=start_timestamp - max_lookback_seconds,
            end=end_timestamp,
            limit=constants.get('global.query-limit'))

        return scheduler_event

    @staticmethod
    def load_wits_stream_event(asset_id, event, max_lookback_seconds):
        """
        To load a wits stream event and get more data if necessary
        :param asset_id: The asset to load
        :param event: The cleaned event from Kafka WITS stream
        :param max_lookback_seconds: Maximum amount of time to look back prior to WITS data to cover gaps
        :return: list of WITS data between the first event timestamp and the first timestamp
        """

        if event and max_lookback_seconds:
            first_timestamp = event[0].get('timestamp')

            if not first_timestamp:
                return event

            # Subtract one from the timestamp so that we don't reselect the final data item that was sent in the event
            end_timestamp = first_timestamp - 1

            previous_events = gather_data_for_period(
                asset_id=asset_id,
                start=first_timestamp - max_lookback_seconds,
                end=end_timestamp,
                limit=constants.get('global.query-limit'))

            event = previous_events + event

        return event

    @staticmethod
    def format_event(event_type: str, event: list) -> list:
        """
        validate the wits_stream event, flatten and organize the data into a desired format
        :param event_type: type of event
        :param event: a wits_stream event
        :return: a list of wits records
        """

        first_event = event[0]
        if isinstance(first_event, list):
            # This is possible for both scheduler and old wits stream event types
            try:
                # Because the events are structured into 'list of lists', using the following code
                # converts them into a flat single list. This is done for better data processing in
                # the proceeding steps.
                event = [item for sublist in event for item in sublist]
            except ValueError:
                raise Exception("Records are not valid JSON")

            if not isinstance(event, list):
                raise Exception("Records is not an array")

            if not event:
                raise Exception("Records is empty.")

            if event_type == 'scheduler':
                event = [
                    {
                        "asset_id": item["asset_id"],
                        "timestamp": int(item['schedule_start'] / 1000)
                    }
                    for item in event
                ]
            return event

        elif isinstance(first_event, dict):
            # This is possible for the new wits stream event type
            records = []
            for each in event:
                records = records + each.get("records", [])
            return records

        else:
            raise TypeError("Event is neither a scheduler, old nor new kafka consumer")

    @staticmethod
    def determine_asset_id(event: list) -> int:
        try:
            return int(event[0]["asset_id"])
        except Exception:
            raise Exception('Event does not contain asset_id: {}'.format(event))

    def load_state(self, state_key: [str, None] = None, auto_migrate: bool = False, raise_warnings: bool = False) -> dict:
        previous_state = super().load_state(state_key=state_key, auto_migrate=auto_migrate, raise_warnings=raise_warnings)

        state = State(self.app_state_fields, previous_state)

        if not state.get('asset_id', None):
            state['asset_id'] = self.asset_id

        return state

    def save_state(self, state_key: [str, None] = None, timestamp: [int, None] = None, raise_warnings: bool = True) -> None:
        if not timestamp:
            timestamp = self.state.get('last_processed_timestamp', 0)

        super().save_state(state_key=state_key, timestamp=timestamp, raise_warnings=raise_warnings)

    def get_modules(self):
        raise exceptions.Misconfigured("No modules found")

    def run_modules(self):
        if not self.event:
            return

        for module_type in self.get_modules():
            try:
                module = module_type(self.state, rollbar=self.rollbar)
            except Exception:
                raise exceptions.Misconfigured("Module {0} not able to initialize for asset_id {1}".format(module_type, self.asset_id))

            try:
                module.run(self.event)
            except Exception:
                message = f"Error in module {module_type.module_key}"
                self.track_error(message=message)
                raise

        last_processed_timestamp = self.event[-1].get('timestamp')
        self.state['last_processed_timestamp'] = last_processed_timestamp



# this constants module can be overridden by your new constants module by importing
# the this module and updating it like this:
#     from worker import constants
#     constants.update({
#         ...
#     })
#
# however, if you have multiple apps and constants in your repository you need to
# make a deep copy (reload) the module in each module to avoid side effects from
# updating the single global constants module. Instead use this:
#     import importlib
#     import worker
#     importlib.reload(worker)
#     from worker import constants
#     constants.update({
#         ...
#     })


from worker.data.operations import get_data_by_path

parameters = {
    "global": {
        "app-name": "my-app",
        "app-key": "my-app-key",
        "event-type": "scheduler",
        "query-limit": 1000,
    },
}


def get(path, default=None):
    return get_data_by_path(data=parameters, path=path, default=default)


def update(additional_parameters):
    """
    The purpose of this method is to update the existing parameters data
    with the provided additional parameters. Note that the globals will
    be replaced if the additional parameters contains global node.
    :param additional_parameters:
    :return:
    """
    globals = parameters.pop('global')
    additional_globals = additional_parameters.pop('global', {})
    globals.update(additional_globals)
    parameters.update(additional_parameters)
    if globals:
        parameters['global'] = globals



import json


class State(dict):

    def __init__(self, fields: dict, state: dict = None, *args, **kwargs):
        """
        :param fields: is a dictionary that holds the state items. The value of each key is the data type.
        :param state:
        :param args:
        :param kwargs:
        """
        self.fields = fields

        super().__init__(*args, **kwargs)
        if not state:
            state = {}

        for field, conversion in self.fields.items():
            self.set_field_if_exists(field, state, conversion)

    def set_field_if_exists(self, field, state, conversion):
        if not field:
            return

        value = None

        if field in state and state[field] is not None:
            value = state[field]
            if conversion:
                value = conversion(value)

        self[field] = value

    def to_json(self):
        output = {field: self.get(field, None) for field in self.fields.keys()}
        output = json.dumps(output)

        return output

    def get(self, key, default=None):
        value = super().get(key, None)
        if value is None:
            value = default
        return value



import numpy as np

from worker.data.operations import nanround
from worker.data.serialization import serialization
from worker.wellbore.model.drillstring_components import Pipe
from worker.wellbore.model.element import Element
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole_section import HoleSection


@serialization
class Ann(Element):
    """
    This class is extended to create annulus segment
    It can be done by passing arguments or pipe and hole section
    """
    SERIALIZED_VARIABLES = {
        'top_depth': float,
        'bottom_depth': float,
        'length': float,

        'inner_diameter_hole': float,
        'outer_diameter_pipe': float,
        'outer_diameter_tooljoint': float,
        'tool_joint_length_ratio': float,
        'hole_type': HoleType
    }

    def __init__(self, **kwargs):
        super(Ann, self).__init__(**kwargs)

        for key in ['inner_diameter_hole', 'outer_diameter_pipe', 'outer_diameter_tooljoint',
                    'tool_joint_length_ratio', 'hole_type']:
            setattr(self, key, kwargs.get(key))

    def set_properties_from_pipe_section(self, pipe: Pipe):
        self.outer_diameter_pipe = pipe.outer_diameter
        self.outer_diameter_tooljoint = pipe.outer_diameter_tooljoint
        self.tool_joint_length_ratio = pipe.tool_joint_length_ratio

    def set_properties_from_hole(self, hole_section: HoleSection):
        self.inner_diameter_hole = hole_section.inner_diameter
        self.hole_type = hole_section.hole_type

    def compute_get_area_pipe_body(self) -> float:
        """
        Compute the annulus flow area consisting of the hole section and pipe body
        :return: area in INCH^2
        """
        return np.pi / 4 * (self.inner_diameter_hole ** 2 - self.outer_diameter_pipe ** 2)

    def compute_get_area_pipe_tooljoint(self) -> float:
        """
        Compute the annulus flow area consisting of the hole section and pipe tooljoint
        :return: area in INCH^2
        """
        return np.pi / 4 * (self.inner_diameter_hole ** 2 - self.outer_diameter_tooljoint ** 2)

    # override
    def compute_get_area(self) -> float:
        """
        Compute the total area of the hole section and return the result
        :return: cross sectional area of the hole section in INCH^2
        """
        area = self.compute_get_area_pipe_body()
        r = self.tool_joint_length_ratio
        if r > 0:
            area = r * self.compute_get_area_pipe_tooljoint() + (1 - r) * area
        return area

    # override
    def compute_get_volume(self):
        """
        This method computes and returns the total volume of the annulus section
        :return: volume in FOOT^3
        """
        area = self.compute_get_area()  # in INCH^2
        volume = area / 144 * self.length  # in FOOT^3
        return volume

    def eq_without_length(self, other):
        if not isinstance(other, Ann):
            return False

        if self.inner_diameter_hole != other.inner_diameter_hole:
            return False
        if self.outer_diameter_pipe != other.outer_diameter_pipe:
            return False
        if self.outer_diameter_tooljoint != other.outer_diameter_tooljoint:
            return False
        if self.tool_joint_length_ratio != other.tool_joint_length_ratio:
            return False
        if self.hole_type != other.hole_type:
            return False

        return True

    def __repr__(self):
        return (
            f"{nanround(self.inner_diameter_hole, 2):>6} in / {nanround(self.outer_diameter_pipe, 2):>6} in, "
            f"{super().__repr__()}"
        )



from typing import Union

from worker.data.operations import get_data_by_path, get_drillstring_by_id

from worker import API
from worker.exceptions import MissingConfigError

from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection
from worker.wellbore.wellbore import Wellbore


def set_drillstring_and_create_wellbore(wits: dict) -> Union[Wellbore, None]:
    """
    Get the wits record and create a wellbore
    :param wits: wits records in json format
    :return: a wellbore
    """
    if wits is None:
        return None

    api_worker = API()

    drillstring_id = get_data_by_path(wits, "metadata.drillstring", str)

    if not drillstring_id:
        return None

    drillstring = get_drillstring_by_id(drillstring_id)

    if not drillstring:
        return None
    active_drillstring_number = drillstring.get("data", {}).get("id")

    # setting active drillstring
    drillstring = Drillstring(drillstring)

    bit = drillstring.get_bit()
    if not bit:
        raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

    asset_id = wits['asset_id']
    bit_size = bit.size

    bit_depth = get_data_by_path(wits, "data.bit_depth", float)
    hole_depth = get_data_by_path(wits, "data.hole_depth", float)

    # setting cased-hole sections
    query = "{data.inner_diameter#gte#%s}" % bit_size
    casings = api_worker.get(
        path="/v1/data/corva", collection="data.casing", asset_id=asset_id, query=query, limit=100
    ).data

    hole = Hole()
    hole.set_casings(casings)
    smallest_cased_hole_diameter = hole.get_min_inner_diameter() or 100

    # ====== setting open-hole sections
    query = "{data.id#lte#%s}" % active_drillstring_number
    sort = "{data.id:1}"
    drillstrings = api_worker.get(
        path="/v1/data/corva", collection="data.drillstring", asset_id=asset_id, query=query, sort=sort, limit=100
    ).data

    # setting open-hole sections
    for ds in drillstrings:
        bit = Drillstring(ds).get_bit()
        if not bit:
            drillstring_id = ds.get("_id")
            raise MissingConfigError(f"Bit was not found in the drillstring with _id='{drillstring_id}'")

        bit_size = bit.size
        # only keeping the DSs that ran below the smallest casing
        if bit_size > smallest_cased_hole_diameter:
            continue

        start_depth = ds.get("data", {}).get("start_depth")

        # if due to the side track the hole depth reduced, then it should update the setting depths in the open hole
        hole.sections = [
            sec for sec in hole.sections
            if sec.hole_type != HoleType.OPEN_HOLE or sec.top_depth <= hole_depth
        ]

        # if a new open-hole section is created in addition to the previous one,
        # the previous bottom depth may need to be modified.
        previous_hole_section = hole.get_last_hole_section()

        top_depth = hole.get_bottom_depth()
        bottom_depth = top_depth + 0.0001
        current_hole_section = HoleSection(
            top_depth=top_depth,
            bottom_depth=bottom_depth,
            inner_diameter=bit_size,
            hole_type=HoleType.OPEN_HOLE
        )

        is_equal_previous_section = False
        if previous_hole_section:
            # The casing bottom depth can't be modified if the difference is more than 300 ft
            cased_hole_condition = previous_hole_section.hole_type == HoleType.CASED_HOLE and start_depth - previous_hole_section.bottom_depth < 300
            if cased_hole_condition or previous_hole_section.hole_type == HoleType.OPEN_HOLE:
                previous_hole_section.set_bottom_depth(start_depth)
                previous_hole_section.set_length()

            is_equal_previous_section = current_hole_section.eq_without_length(previous_hole_section)

        if is_equal_previous_section:
            continue

        current_hole_section.set_length()
        hole.add_section(current_hole_section)

    # In some cases the DS setting depth doesn't match the hole depth so the wits data overrides it;
    # this is only applied to the top depth of the last section and bottom depth of the previous section.
    # Note: only in cases that the wits hole_depth < setting depth
    # Condition: the number of open hole sections should be at least two needs to be satisfied first
    if len(hole) - len(hole.get_cased_hole()) > 1 and 1 < hole_depth < hole[-1].top_depth:
        hole[-2].set_bottom_depth(hole_depth)
        hole[-2].set_length()

        hole[-1].set_top_depth(hole_depth)
        hole[-1].set_length()

    return Wellbore(
        drillstring=drillstring,
        hole=hole,
        bit_depth=bit_depth,
        hole_depth=hole_depth,
    )



import unittest

from worker.test import AppTestRun
from worker import API
from worker.test.lambda_function_test_run import generate_runner_parser
from testcase.app import app_lambda


class TestLambdaApp(unittest.TestCase):

    def test_lambda_app(self):
        asset_id = 3083

        def run():
            collections = ['drilling-efficiency.mse']

            parser = generate_runner_parser()
            args = parser.parse_args([
                '--asset_id', str(asset_id),
                '--start_timestamp', '1502041854',
                '--end_timestamp', '1502043349',
                '--timestep', '600',
                '--to_delete', str(True),
                '--storage_type', str('mongo')
            ])

            # loading of constants should happen here to avoid conflicts
            from testcase.app.app_constants import constants
            app = AppTestRun(app_lambda.lambda_handler, collections, constants, args=args)
            app.run()

        try:
            run()

            api = API()
            records = api.get(
                path="/v1/data/corva", collection='drilling-efficiency.mse',
                asset_id=asset_id, sort="{timestamp: 1}", limit=1000,
            ).data

            self.assertEqual(42, len(records))
        except Exception:
            self.fail("This run should not fail.")



# The modules of this files are used for (de)serialization of any object
#
# The only things you need to do is to add '@serialization' decorator to
# the class. If the class is an enum that's enough. For other class types,
# you need to specify the variables that need to be serialized as a dict:
# variable name as key -> variable type as value
# the name of it should be 'SERIALIZED_VARIABLES'
#
# At the end you can use:
# 'obj2json' to serialize, and
# 'json2obj' to deserialize
#


import json
from enum import Enum
from typing import Union

from worker.data.two_way_dict import TwoWayDict

# This registry object keeps a list of classes that subscribed to serialization.
registry = TwoWayDict()


# This method is used as a decorator of the class that wants to be serialized
def serialization(cls):
    registry[cls] = f"_{cls.__name__.lower()}_"
    return cls


class Encoder(json.JSONEncoder):
    """
    This class is used to encode: object --> json string
    """

    def default(self, obj):
        cls = obj.__class__

        if cls in registry:
            cls_tag = registry[cls]
            d = {cls_tag: True}

            # enum types
            if issubclass(cls, Enum):
                d['name'] = obj.name
                return d

            # class instances
            # 1. classes with serialize method
            if hasattr(obj, 'serialize'):
                d.update(obj.serialize())
                return d

            # 2. classes without it
            for var_name in cls.SERIALIZED_VARIABLES.keys():
                value = getattr(obj, var_name)
                d[var_name] = value
            return d

        return json.JSONEncoder.default(self, obj)


def obj2json(obj: object, output_format=str) -> Union[str, dict]:
    """
    Convert an object to JSON. The object class should have been decorated
    with serialization function.
    :param obj:
    :param output_format: str, or dict
    :return:
    """
    output = json.dumps(obj, cls=Encoder)
    if output_format == str:
        return output
    if output_format == dict:
        return json.loads(output)
    raise TypeError("output_format should be str or dict.")


def _set_from_json(cls, data: str):
    if isinstance(data, str):
        data = json.loads(data)

    # getitem of an enum is the best way of de-serializing it
    if issubclass(cls, Enum):
        return cls[data['name']]

    return cls(**data)


def _json_to_obj_hook(j_str):
    cls_tags = [tag for tag in registry.values() if isinstance(tag, str)]
    for tag in cls_tags:
        if tag in j_str:
            cls = registry.get(tag)
            return _set_from_json(cls, j_str)

    return j_str


def json2obj(json_str: Union[str, dict]):
    if json_str is None:
        return None

    if not isinstance(json_str, (str, dict)):
        raise TypeError(f"Wrong json format: {type(json_str)}. Acceptable formats are str and dict.")

    if isinstance(json_str, dict):
        json_str = json.dumps(json_str)

    return json.loads(json_str, object_hook=_json_to_obj_hook)



class CorvaException(Exception):
    pass


class Misconfigured(Exception):
    pass


class APIError(CorvaException):
    pass


class NotFound(CorvaException):
    pass


class Forbidden(CorvaException):
    pass


class MissingConfigError(Exception):
    pass



import argparse
import json
import os

from cached_property import cached_property
from tqdm import tqdm

from worker.data import operations
from worker.mixins.logging import LoggingMixin


API_DATA_PATH = '/v1/data/corva/'
API_ASSET_PATH = '/v1/assets/'


def generate_transfer_parser():
    """
    Creating the supporting arguments
    :return:
    """
    parser = argparse.ArgumentParser(description="Run your tests locally on an asset.")
    parser.add_argument("-v", "--source_environment", "--env", type=str, required=True,
                        help="source environment, options: 'qa', 'staging', 'production'")
    parser.add_argument("-a", "--source_asset_id", "--id", type=int, required=True,
                        help="set source asset_id")
    parser.add_argument("-s", "--start_timestamp", "--start", type=int, required=False, default=None,
                        help="start timestamp for the main collections to be copied over")
    parser.add_argument("-e", "--end_timestamp", "--end", type=int, required=False, default=None,
                        help="end timestamp for the main collections to be copied over")
    parser.add_argument("-c", "--config_collections", "--config", type=json.loads, required=False, default=None,
                        help="setting a new config collection list that override the existing one")
    parser.add_argument("-m", "--main_collections", "--main", type=json.loads, required=False, default=None,
                        help="setting a new main collection list that override the existing one")
    return parser


class ToLocalTransfer(LoggingMixin):
    # config collection: this can be modified in __init__
    CONFIG_COLLECTIONS = [
        "data.surface-equipment",
        "data.actual_survey",
        "data.plan_survey",

        # in the following collections the setting timestamps need to be removed
        # but for the purpose of this test analysis is not important.
        "data.casing",
        "data.drillstring",
        "data.mud"

        # not required for most of the cases
        # "data.formations",
        # "data.pressure-gradient",
        # "data.npt-events",
        # "data.costs",
        # "data.well-sections",
    ]

    # collections other than configs: this can be modified in __init__
    MAIN_COLLECTIONS = [
        "wits"
    ]

    def __init__(self, args):
        super().__init__()

        self.source_env = None
        self.source_asset_id = None
        self.source_company_id = None

        self.local_env = "local"
        self.local_asset_id = None

        # this is used to avoid copying configs if it is not a new well
        self.is_new_well = True

        self.app_name = os.getenv("APP_NAME", "LocalTesting")

        # The purpose of this variable is to change the config _ids
        # For instance when the wits records are copied over to local
        # the drillstring node under metadata is still referring to source
        # _id so keeping a mapping is required to replace the _ids.
        # In addition this variable is stored on local redis.
        self.config_id_mapper = {}

        self.start_timestamp = None
        self.end_timestamp = None

        if args is None:
            parser = generate_transfer_parser()
            args = parser.parse_args()

        self.initialize(args)

    def initialize(self, args):
        if args.config_collections:
            self.CONFIG_COLLECTIONS = args.config_collections

        if args.main_collections:
            self.MAIN_COLLECTIONS = args.main_collections

        self.source_env = args.source_environment
        self.source_asset_id = args.source_asset_id

        self.start_timestamp = args.start_timestamp
        self.end_timestamp = args.end_timestamp

    def run(self):
        self.setup_local_well()
        self.store_config_data()
        self.store_other_collections()

    @cached_property
    def source_api_worker(self):
        return operations.setup_api_worker(self.source_env, self.app_name)

    @cached_property
    def local_api_worker(self):
        return operations.setup_api_worker(self.local_env, self.app_name)

    @cached_property
    def local_redis_worker(self):
        return operations.setup_redis_worker(self.local_env)

    def _get_well_properties(self):
        path = "%s?ids[]=%s" % (API_ASSET_PATH, self.source_asset_id)

        well_properties = self.source_api_worker.get(path=path)
        if well_properties.count != 1:
            raise Exception(f"Count mismatch: there should be only one asset! {well_properties.count} found!")

        well_properties = well_properties.data[0]

        if well_properties['asset_type'] != 'well':
            raise Exception("Not a well")

        self.well_name = well_properties['name']
        self.source_company_id = well_properties['company_id']

    def setup_local_well(self):
        """
        Set up the local well. If the well is not already setup, a new one will be created.
        :return:
        """
        self._get_well_properties()

        path = "%s?order=asc&page=1&per_page=100&search=%s&sort=name&types[]=well" % (API_ASSET_PATH, self.well_name)
        wells = self.local_api_worker.get(path=path)

        found = False
        if wells.count >= 1:
            well = wells.data[0]
            well_name2 = well['name']
            if self.well_name == well_name2:
                found = True
                print("Found local asset!")

        if not found:
            print("Well not found! Trying to create a new asset ...")
            data = {
                "name": self.well_name,
                "parent_asset_id": None,
                "company_id": 1,
                "visibility": "visible",
                "asset_type": "well",
                "type": "Asset::Well"
            }
            data_json = json.dumps(data)
            well = self.local_api_worker.post(path=API_ASSET_PATH, data=data_json).data
            print("Well is created!")

        self.is_new_well = not found
        self.local_asset_id = well['id']
        print(f"Local asset_id={self.local_asset_id}")

    def store_config_data(self):
        """
        Store the config collection data into local database
        :return:
        """
        if not self.is_new_well:
            print("Not a new well.")
            config_id_mapper = self.local_redis_worker.get(self.get_config_mapper_redis_key())
            if config_id_mapper:
                self.config_id_mapper = json.loads(config_id_mapper)
                print("Copying configs are skipped.")
                return

        # maps the source to local ids in a dict of dicts in the format {collection:{source_id: local_id}}
        config_id_mapper = {}

        # get records
        print("Copying configs to local ...")
        print(f"{'Collection':30} {'Count':10} {'Post':10} ")
        print(f"{'-' * 30:30} {'-' * 10:10} {'-' * 10:10} ")

        count_records = 0
        for col in self.CONFIG_COLLECTIONS:
            records = []
            print(f"{col:30} ", end='')
            col_data = self.source_api_worker.get(
                path=API_DATA_PATH, collection=col, asset_id=self.source_asset_id, limit=10000
            )
            print(f"{col_data.count:>10} ", end="")
            if col_data.count < 1:
                print()
                continue

            records.extend(col_data.data)

            source_ids = [record['_id'] for record in records]

            for record in records:
                record['company_id'] = 1
                record['asset_id'] = self.local_asset_id
                record.pop('_id', None)

            # post
            results = self.local_api_worker.post(path=API_DATA_PATH, data=json.dumps(records))
            print(f"{'successful':10} ", end="\n")

            local_ids = results.data['ids']
            count_records += len(local_ids)

            if len(source_ids) != len(local_ids):
                raise IndexError(f"Number of source ids ({len(source_ids)}) does not match local ids ({len(local_ids)})")

            collection_id_mapper = {source_ids[i]: local_ids[i] for i in range(len(source_ids))}
            config_id_mapper[col] = collection_id_mapper

        print(f"{count_records} records are posted!")
        self.debug(self.local_asset_id, f"id Mapper: {config_id_mapper}")

        self.config_id_mapper = config_id_mapper
        self.local_redis_worker.set(self.get_config_mapper_redis_key(), json.dumps(self.config_id_mapper))

    def store_other_collections(self, show_progress: bool = True):
        """
        Store the other config data on local
        :param show_progress: to show progress bar or not
        :return:
        """
        collections = self.MAIN_COLLECTIONS

        # to remove duplicate collections
        collections = list(set(collections))

        start_timestamp = self.start_timestamp
        end_timestamp = self.end_timestamp

        for col in collections:
            # Finding the start and end if not available.
            if not start_timestamp:
                res = self.source_api_worker.get(
                    path=API_DATA_PATH, collection=col, asset_id=self.source_asset_id, sort="{timestamp:+1}", limit=1
                )

                if res.count < 1:
                    continue
                start_timestamp = res.data[0].get("timestamp")

            if not end_timestamp:
                res = self.source_api_worker.get(
                    path=API_DATA_PATH, collection=col, asset_id=self.source_asset_id, sort="{timestamp:-1}", limit=1
                )
                end_timestamp = res.data[0].get("timestamp")

            res = self.local_api_worker.get(
                path=API_DATA_PATH, collection=col, asset_id=self.local_asset_id, sort="{timestamp:-1}", limit=1
            )
            if res.count > 0:
                ts = res.data[0].get('timestamp', 0)
            else:
                ts = start_timestamp - 1

            print(f"'{col}' collection, Source Range: [{start_timestamp}, {end_timestamp}], Local Start: {ts}")

            limit = 3600 if col == 'wits' else 100

            pbar = tqdm(total=100, ncols=100, disable=not show_progress)

            proceed = True
            while proceed:
                query = "{timestamp#gt#%s}AND{timestamp#lte#%s}" % (ts, end_timestamp)

                res = self.source_api_worker.get(
                    path=API_DATA_PATH, collection=col, asset_id=self.source_asset_id,
                    query=query, sort="{timestamp:+1}", limit=limit
                )

                count = res.count
                if count < 1:
                    break

                records = res.data

                data_str = json.dumps(records)
                data_str = data_str.replace(
                    "\"company_id\": %s" % self.source_company_id,
                    "\"company_id\": %s" % 1
                )
                data_str = data_str.replace(
                    "\"asset_id\": %s" % self.source_asset_id,
                    "\"asset_id\": %s" % self.local_asset_id
                )
                if col == 'wits':
                    for config_name, mapped_ids in self.config_id_mapper.items():
                        for id_source, id_local in mapped_ids.items():
                            data_str = data_str.replace(id_source, id_local)

                records = json.loads(data_str)

                for record in records:
                    record.pop('_id', None)

                res = self.local_api_worker.post(path=API_DATA_PATH, data=json.dumps(records))

                pbar_added = round((records[-1]['timestamp'] - ts) / (end_timestamp - start_timestamp) * 100)
                pbar.update(pbar_added)

                ts = records[-1]['timestamp']

            pbar.update(100 - pbar.n)

    def get_config_mapper_redis_key(self):
        return "%s-config_id_mapper" % self.local_asset_id



from enum import Enum

from worker.data.serialization import serialization
from worker.data.unit_conversions import KG_M3_to_PPG


@serialization
class HoleType(Enum):
    CASED_HOLE = "Cased Hole"
    OPEN_HOLE = "Open Hole"


@serialization
class PipeType(Enum):
    Unknown = "Unknown"
    DP = "Drillpipe"
    HWDP = "Heavy Weight Drillpipe"
    DC = "Drill Collar"
    STABILIZER = "Stabilizer"
    AGITATOR = "Agitator"
    PDM = "Positive Displacement Motor"
    MWD = "Measurement While Drilling"
    RSS = "Rotary Steerable System"
    BIT = "Bit"

    @staticmethod
    def determine_type(family: str):
        family = family.lower()
        mapping = {
            'dp': PipeType.DP,
            'hwdp': PipeType.HWDP,
            'dc': PipeType.DC,
            'stabilizer': PipeType.STABILIZER,
            'agitator': PipeType.AGITATOR,
            'pdm': PipeType.PDM,
            'mwd': PipeType.MWD,
            'rss': PipeType.RSS,
            'bit': PipeType.BIT
        }
        _type = mapping.get(family)
        if not _type:
            _type = PipeType.Unknown

        return _type


@serialization
class PipeMaterial(Enum):
    Steel = {
        'density': 7850 * KG_M3_to_PPG,
        'modulus_of_elasticity': 29.9e6,
        'poisson_ratio': 0.30
    }
    Aluminium = {
        'density': 2700 * KG_M3_to_PPG,
        'modulus_of_elasticity': 10.5e6,
        'poisson_ratio': 0.32
    }
    Titanium = {
        'density': 4506 * KG_M3_to_PPG,
        'modulus_of_elasticity': 16.5e6,
        'poisson_ratio': 0.31
    }
    NonMagnetic = {
        'density': 7850 * KG_M3_to_PPG,
        'modulus_of_elasticity': 29.9e6,
        'poisson_ratio': 0.30
    }

    def get_density(self) -> float:
        """
        Get the density of the pipe material
        :return: pipe density PPG
        """
        return self['density']

    def get_modulus_of_elasticity(self) -> float:
        """
        Get the pipe modulus of elasticity
        :return: modulus of elasticity in PSI
        """
        return self['modulus_of_elasticity']

    def get_poisson_ratio(self) -> float:
        """
        Get the Poisson ratio
        :return: Poisson ratio
        """
        return self['poisson_ratio']



from worker.app.modules.__init__ import Module
from worker.framework import constants


class Scheduler(Module):
    """
    This is an abstract module that needs to be extended by actual modules.
    An example of this module would be running an app at the frequency of
    once every X seconds regardless of the activity.
    """
    module_key = "scheduler_module"
    collection = "scheduler_collection"
    module_state_fields = {
        'last_processed_timestamp': int
    }

    # override
    def run(self, wits_stream: list):
        super().run(wits_stream)

        dataset = self.load_dataset(wits_stream)

        results = []
        for el in dataset:
            if self.should_run_processor(el):
                results = self.run_module(el, results)

        self.save_state()
        self.store_output(self.global_state['asset_id'], results)

    # override
    def should_run_processor(self, event: dict):
        """
        check one wits event and test against the last processed timestamp
        :param event:
        :return:
        """
        timestamp = event.get('timestamp', 0)
        running_frequency = constants.get("{}.{}.running_frequency".format(self.app_key, self.module_key))
        if self.state.get('last_processed_timestamp', 0) + running_frequency <= timestamp:
            # update the last_processed_timestamp
            self.set_state('last_processed_timestamp', timestamp)
            return True

        return False



import json

from typing import Union

from worker.data.api import API
from worker.framework import constants
from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state.mixins import StateMixin

from worker.state.state import State


class Module(StateMixin, LoggingMixin, RollbarMixin):
    """
    This is an abstract base module that needs to be extended by an actual module.
    """
    # module_key is used for redis access and state of this module
    module_key = "module"
    collection = "collection"
    module_state_fields = {
        'last_processed_timestamp': int
    }

    def __init__(self, global_state, *args, **kwargs):
        self.app_key = constants.get('global.app-key')
        self.app_name = constants.get('global.app-name')
        self.asset_id = global_state.get("asset_id")

        state_storage_type = constants.get("global.state_storage_type", None)
        if state_storage_type:
            kwargs["state_storage_type"] = state_storage_type

        self.global_state = global_state

        super().__init__(*args, **kwargs)

    def run(self, wits_stream: list):
        """
        :param wits_stream: a wits stream event
        :return:
        """
        # load the state
        state = self.load_state()
        self.state = self.process_module_state(state)

        # subclasses should implement their own run

    def should_run_processor(self, event):
        raise Exception("This method need to be implemented by subclasses!")

    def load_state(self, state_key: [str, None] = None, auto_migrate: bool = False, raise_warnings: bool = False) -> dict:
        current_state = super().load_state(state_key=state_key, auto_migrate=auto_migrate, raise_warnings=raise_warnings)
        return State(self.module_state_fields, current_state)

    @staticmethod
    def process_module_state(state):
        return state

    def set_state(self, key, value):
        self.state[key] = value

    def save_state(self, state_key: [str, None] = None, timestamp: [int, None] = None, raise_warnings: bool = True) -> None:
        if not timestamp:
            timestamp = self.global_state.get('last_processed_timestamp', 0)

        super().save_state(state_key=state_key, timestamp=timestamp, raise_warnings=raise_warnings)

    def load_dataset(self, event):
        return event

    def run_module(self, dataset: Union[list, dict], beginning_results: list) -> list:
        raise Exception("Not implemented")

    def get_last_exported_timestamp_from_collection(self, asset_id, query=None, less_than=None):
        """
        Query the module collection for this asset_id + module, sorted by timestamp descending,
        limit 1, grab the last item's timestamp. Default to 0 if no records found.
        @asset_id:
        @less_than: the timestamp before which you want to get
        """
        if less_than:
            query = query or ""
            query += "AND{timestamp#lt#%s}" % less_than

        worker = API()
        last_document = worker.get(
            path="/v1/data/corva", query=query, collection=self.collection, asset_id=asset_id,
            sort="{timestamp: -1}", limit=1,
        ).data

        if not last_document:
            return 0

        last_document = last_document[0]
        last_processed_timestamp = last_document.get('timestamp', 0)

        return last_processed_timestamp

    @staticmethod
    def gather_first_wits_timestamp_since(asset_id: int, since: int, activity_fields=None, operator='eq') -> int:
        """
        Query the Wits collection for this asset_id where state in wits_states and timestamp >= since
        """

        query = '{timestamp#%s#%s}' % ('gt', since)

        operator = operator.lower()

        if activity_fields:
            if operator == "eq" and isinstance(activity_fields, list):
                operator = "in"

            if operator in ("in", "nin"):
                if not isinstance(activity_fields, list):
                    activity_fields = [activity_fields]

                # Put each state into a formatted string for querying
                activity_fields = ["'{0}'".format(state) for state in activity_fields]
                activity_fields = "[{0}]".format(",".join(activity_fields))
            else:
                activity_fields = "'{0}'".format(activity_fields)

            query += 'AND{data.state#%s#%s}' % (operator, activity_fields)

        worker = API()
        first_wits_since = worker.get(
            path="/v1/data/corva", collection='wits', asset_id=asset_id, sort="{timestamp: 1}", limit=1, query=query
        ).data

        if not first_wits_since:
            return 0

        first_wits_since = first_wits_since[0]
        first_wits_since_timestamp = first_wits_since.get('timestamp', 0)

        return first_wits_since_timestamp

    @staticmethod
    def gather_maximum_timestamp(event, start, activity_fields):
        """
        get the maximum time stamp of a stream of data
        :param event: a stream of data  that the majority is wits collection
        :param start:
        :param activity_fields:
        :return:
        """
        maximum_timestamp = start
        for data in event:
            if data.get("collection") == "wits" and data.get('data', {}).get('state', None) in activity_fields:
                maximum_timestamp = max(data.get("timestamp", 0), maximum_timestamp)

        return maximum_timestamp

    def gather_minimum_timestamp(self, asset_id: int, event: list):
        minimum = self.get_last_exported_timestamp_from_collection(asset_id)

        if not minimum:
            minimum = event[0]["timestamp"] - 1800

        return minimum

    def gather_collections_for_period(self, asset_id, start, end, query=None):
        limit = constants.get("global.query-limit")

        query = query or ""
        if query:
            query += "AND"

        query += "{timestamp#gte#%s}AND{timestamp#lte#%s}" % (start, end)

        worker = API()
        dataset = worker.get(
            path="/v1/data/corva", collection=self.collection, asset_id=asset_id, query=query,
            sort="{timestamp: 1}", limit=limit,
        ).data

        if not dataset:
            return []

        return dataset

    def store_output(self, asset_id, output):
        """
        to store/post results
        :param asset_id: asset id of the well
        :param output: an array of json objects to be posted
        :return: None
        """

        if not asset_id or not output or not self.collection:
            return

        output = self.format_output(output)

        self.debug(asset_id, "{0} output -> {1}".format(self.module_key, output))

        worker = API()
        worker.post(path="/v1/data/corva", data=output)

    def build_empty_output(self, wits: dict) -> dict:
        """
        Building an empty output result.
        :param wits: one wits record
        :return:
        """
        output = {
            'timestamp': int(wits.get('timestamp')),
            'company_id': int(wits.get('company_id')),
            'asset_id': int(wits.get('asset_id')),
            'provider': str(wits.get('provider', 'corva')),
            'version': 1,
            'collection': self.collection,
            'data': {}
        }
        return output

    @staticmethod
    def format_output(output):
        output = json.dumps(output)
        return output

    @staticmethod
    def compute_timestep(dataset: list):
        if len(dataset) < 2:
            return None

        timestep = (dataset[-1].get('timestamp') - dataset[0].get('timestamp')) / (len(dataset) - 1)

        return timestep



from testcase.state_testing_app.state_testing_app import TestingApp


def lambda_handler(event, context):
    """
    This function is the main entry point of the AWS Lambda function
    :param event: a scheduler or kafka event
    :param context: AWS Context
    :return:
    """
    app = TestingApp()
    app.load(event)
    app.asset_id = event[0].get("asset_id")
    app.state = app.load_state()

    app.state["last_processed_timestamp"] = 0
    app.state["message"] = f"This is the app state on {app.state_storage_type.value}"
    app.event = event
    app.run_modules()
    app.save_state()



import numpy as np

from worker.data.operations import nanround, equal
from worker.data.serialization import serialization

from worker.wellbore.model.element import Element
from worker.wellbore.model.enums import HoleType


@serialization
class HoleSection(Element):
    SERIALIZED_VARIABLES = {
        'top_depth': float,
        'bottom_depth': float,
        'length': float,

        'inner_diameter': float,
        'hole_type': HoleType
    }

    def __init__(self, **kwargs):
        super(HoleSection, self).__init__(**kwargs)
        self.inner_diameter = kwargs['inner_diameter']
        self.hole_type = kwargs.get('hole_type')

    def __eq__(self, other):
        params = [
            'inner_diameter',
            'length',
            'hole_type'
        ]
        return equal(self, other, params)

    def eq_without_length(self, other):
        params = [
            'inner_diameter',
            'hole_type'
        ]
        return equal(self, other, params)

    # override
    def compute_get_area(self) -> float:
        """
        Compute the total area of the hole section and return the result
        :return: cross sectional area of the hole section in INCH^2
        """
        return np.pi / 4 * self.inner_diameter ** 2

    # override
    def compute_get_volume(self) -> float:
        """
        Compute the volume of the hole section
        :return: volume of the hole section in FOOT^3
        """
        volume = self.compute_get_area() * self.length  # in INCH^2 * FOOT
        volume /= 144  # in FOOT^3
        return volume

    def __repr__(self):
        return (
            f"{nanround(self.inner_diameter, 2):>6} in, "
            f"{super().__repr__()}, "
            f"{self.hole_type.value}"
        )



from testcase.app.surface_mse_module import SurfaceMSE
from worker import App


class DrillingEfficiency(App):

    # override
    def get_modules(self):
        return [
            SurfaceMSE
        ]



from typing import List


def get_unique_measured_depths(*items) -> List:
    """
    Find the edge measure depths of the given items
    :param items: an object of class Drilling or Hole
    :return:
    """
    mds = {depth for item in items for sec in item for depth in (sec.top_depth, sec.bottom_depth)}
    mds = sorted(list(mds))  # sorting them
    return mds



import os
import unittest

from worker.data.operations import get_data_by_path, get_one_data_record
from worker.data.serialization import obj2json, json2obj
from worker.test.utils import file_to_json
from worker.wellbore.factory import set_drillstring_and_create_wellbore
from worker.wellbore.measured_depth_finder import get_unique_measured_depths
from worker.wellbore.model.annulus import Annulus
from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import HoleType, PipeType
from worker.wellbore.model.hole import Hole
from worker.wellbore.model.hole_section import HoleSection


class TestSections(unittest.TestCase):
    def test_hole_section(self):
        hole_section1 = HoleSection(inner_diameter=5, top_depth=0, bottom_depth=20_000, hole_type=HoleType.OPEN_HOLE)

        hole_section1.set_length()

        self.assertEqual(20_000, hole_section1.length)

        expected_area = 19.634954084936208  # in INCH^2
        expected_volume = 2727.07695624114  # in FOOT^3

        self.assertAlmostEqual(expected_area, hole_section1.compute_get_area(), places=10)
        self.assertAlmostEqual(expected_volume, hole_section1.compute_get_volume(), places=10)

    def test_hole(self):
        casings = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "casings.json")
        )
        casings = file_to_json(casings)

        hole = Hole()
        hole.set_casings(casings)
        self.assertEqual(2, len(hole))

        hole_section1 = hole[0]
        self.assertAlmostEqual(6.88, hole_section1.inner_diameter, 10)
        self.assertAlmostEqual(6_770, hole_section1.length, 10)
        self.assertAlmostEqual(0, hole_section1.top_depth, 10)
        self.assertAlmostEqual(6_770, hole_section1.bottom_depth, 10)
        self.assertEqual(HoleType.CASED_HOLE, hole_section1.hole_type)

        hole_section2 = hole[1]
        self.assertAlmostEqual(4.67, hole_section2.inner_diameter, 10)
        self.assertAlmostEqual(14192, hole_section2.length, 10)
        self.assertAlmostEqual(6_770, hole_section2.top_depth, 10)
        self.assertAlmostEqual(20_962, hole_section2.bottom_depth, 10)
        self.assertEqual(HoleType.CASED_HOLE, hole_section2.hole_type)

        self.assertAlmostEqual(20_962, hole.get_bottom_depth(), 10)

        # Open hole
        open_hole = HoleSection(inner_diameter=4.5, top_depth=1111, bottom_depth=21_000, hole_type=HoleType.OPEN_HOLE)
        hole.add_section(open_hole)
        hole_section3 = hole[2]
        self.assertAlmostEqual(4.5, hole_section3.inner_diameter, 10)
        self.assertAlmostEqual(21_000 - 20_962, hole_section3.length, 10)
        self.assertAlmostEqual(20_962, hole_section3.top_depth, 10)
        self.assertAlmostEqual(21_000, hole_section3.bottom_depth, 10)
        self.assertEqual(HoleType.OPEN_HOLE, hole_section3.hole_type)

        self.assertAlmostEqual(21_000, hole.get_bottom_depth(), 10)

        # test fetching of the correct casing at a given depth
        casing1 = hole.find_section_at_measured_depth(10_000)
        self.assertEqual(id(hole_section2), id(casing1))

        # testing unique boundary depths
        mds = get_unique_measured_depths(hole)
        self.assertListEqual(mds, [0, 6770.0, 20962.0, 21000.0])

        # test serialization
        ser_str = obj2json(hole)
        hole_des = json2obj(ser_str)
        self.assertEqual(len(hole), len(hole_des))
        self.assertAlmostEqual(hole_des[1].inner_diameter, hole[1].inner_diameter, 10)
        self.assertAlmostEqual(hole_des[1].top_depth, hole[1].top_depth, 10)

        print("Hole:")
        print(hole)
        return hole

    def test_drillstring(self):
        drillstring_json_file = os.path.abspath(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "drillstring.json")
        )
        drillstring_json = file_to_json(drillstring_json_file)

        drillstring = Drillstring()
        drillstring.set_drillstring(drillstring_json)

        component_1 = drillstring[0]
        self.assertAlmostEqual(3.826, component_1.inner_diameter, 10)
        self.assertAlmostEqual(4.5, component_1.outer_diameter, 10)
        self.assertAlmostEqual(31, component_1.joint_length, 10)
        self.assertAlmostEqual(31, component_1.length, 10)

        component_6 = drillstring[5]
        self.assertAlmostEqual(2.75, component_6.inner_diameter, 10)
        self.assertAlmostEqual(4.75, component_6.outer_diameter, 10)
        self.assertAlmostEqual(29.09, component_6.joint_length, 10)
        self.assertAlmostEqual(29.09, component_6.length, 10)

        bit = drillstring[-1]
        self.assertAlmostEqual(6.75, bit.size, 10)
        self.assertAlmostEqual(1.1780972450961724, bit.tfa, 10)
        self.assertEqual(PipeType.BIT.value, bit.pipe_type.value, 10)

        # test for adjusting depths using bit depth
        wits = get_wits()
        drillstring.update(wits)
        self.assertAlmostEqual(wits.get('data', {}).get('bit_depth'), drillstring[-1].bottom_depth, 10)

        # test for finding drillstring section at a measured depth
        drillstring_section3 = drillstring.find_section_at_measured_depth(8_000)
        self.assertAlmostEqual(3.83, drillstring_section3.inner_diameter, 1)
        self.assertAlmostEqual(4.5, drillstring_section3.outer_diameter, 10)

        print("Drillstring")
        print(drillstring)
        return drillstring

    def test_annulus(self):
        drillstring = self.test_drillstring()
        hole = self.test_hole()
        annulus = Annulus(drillstring=drillstring, hole=hole)
        print("Annulus:")
        print(annulus)

        # 2nd DP should be split into two segments
        section2, section3 = annulus[2: 4]
        self.assertAlmostEqual(6.88, section2.inner_diameter_hole, 10)
        self.assertEqual(HoleType.CASED_HOLE, section2.hole_type)
        self.assertAlmostEqual(4.67, section3.inner_diameter_hole, 10)
        self.assertEqual(HoleType.CASED_HOLE, section3.hole_type)

        # the segment above the is extended to the bit depth
        last_section = annulus[-1]

        component_above_bit, bit = drillstring[-2:]
        combined_length = component_above_bit.length + bit.length

        self.assertAlmostEqual(combined_length, last_section.length, 10)

    def test_wellbore_creation_with_factory(self):
        """
        In this test case the creation of the wellbore through
        factory is shown; note that only the wits record is used
        and the other elements (casings, drillstrings) are fetched
        from the API.
        :return:
        """
        wits = get_wits()
        bit_depth = get_data_by_path(wits, 'data.bit_depth', float)
        hole_depth = get_data_by_path(wits, 'data.hole_depth', float)
        wellbore = set_drillstring_and_create_wellbore(wits)

        print("\n Hole configuration is:")
        hole = wellbore.actual_hole
        print(hole)
        self.assertEqual(2, len(hole))
        self.assertEqual(HoleType.CASED_HOLE, hole[0].hole_type)
        self.assertEqual(HoleType.OPEN_HOLE, hole[1].hole_type)
        self.assertAlmostEqual(hole_depth, hole.get_bottom_depth(), 10)

        print("\n Drillstring configuration is:")
        drillstring = wellbore.actual_drillstring
        self.assertAlmostEqual(bit_depth, drillstring.bit_depth, 10)
        print(drillstring)

        print("\n Annulus configuration is:")
        annulus = wellbore.actual_annulus
        print(annulus)

        # testing drilling body volume change
        vol_change_calculated = wellbore.compute_get_drillstring_body_volume_change(6000, 5900)
        area_dp = wellbore.actual_drillstring[0].compute_body_cross_sectional_area_tj_adjusted()  # in INCH^2
        vol_expected = area_dp / 144 * -100
        self.assertAlmostEqual(vol_change_calculated, vol_expected, 10)

        # test serialization
        ser_str = obj2json(wellbore)
        wellbore_des = json2obj(ser_str)
        self.assertEqual(len(wellbore.actual_drillstring), len(wellbore_des.actual_drillstring))
        self.assertAlmostEqual(wellbore.actual_drillstring[1].inner_diameter, wellbore_des.actual_drillstring[1].inner_diameter, 10)
        self.assertAlmostEqual(wellbore.actual_drillstring[1].top_depth, wellbore_des.actual_drillstring[1].top_depth, 10)

    def test_overlapping_liners(self):
        """
        In this test case we are checking if the casings and liners
        are set correctly even if the there are overlapping liners.
        :return:
        """

        wits = get_one_data_record(asset_id=3083, collection="wits")
        wellbore = set_drillstring_and_create_wellbore(wits)

        hole: Hole = wellbore.actual_hole
        hole_depth = 25633.01

        self.assertEqual(3, len(hole))

        self.assertEqual(HoleType.CASED_HOLE, hole[0].hole_type)
        self.assertAlmostEqual(0.0, hole[0].top_depth, 1)
        self.assertAlmostEqual(20633.0, hole[0].bottom_depth, 1)
        self.assertAlmostEqual(4.28, hole[0].inner_diameter, 0.01)

        self.assertEqual(HoleType.CASED_HOLE, hole[1].hole_type)
        self.assertAlmostEqual(20633.0, hole[1].top_depth, 1)
        self.assertAlmostEqual(23000.0, hole[1].bottom_depth, 1)
        self.assertAlmostEqual(3.0, hole[1].inner_diameter, 0.01)

        self.assertEqual(HoleType.OPEN_HOLE, hole[2].hole_type)
        self.assertAlmostEqual(23000.0, hole[2].top_depth, 1)
        self.assertAlmostEqual(hole_depth, hole[2].bottom_depth, 1)
        self.assertAlmostEqual(2.85, hole[2].inner_diameter, 0.01)

    def test_creating_wellbore_with_missing_drillstring(self):
        wits_record = {
            "collection": "wits",
            "asset_id": 10477107,
            "metadata": {
                'drillstring': '5f4d5bf1150c40007f3a12fc'
            },
            "data": {
                "bit_depth": 1111,
                "hole_depth": 1111
            }
        }

        wellbore = set_drillstring_and_create_wellbore(wits_record)
        self.assertIsNone(wellbore)


def get_wits():
    wits_json_file = os.path.abspath(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "test", "wits.json")
    )
    return file_to_json(wits_json_file)



import os
from typing import List

from cached_property import cached_property

from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin
from worker.state.enums import StateStorageType
from worker.state.handlers.mongo import MongoStateHandler
from worker.state.handlers.redis import RedisStateHandler


class StateMixin(LoggingMixin, RollbarMixin):

    def __init__(self, state_storage_type=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Get from environment if not provided, default to redis
        if state_storage_type:
            self.state_storage_type = StateStorageType(state_storage_type)  # Raises error if not in StateStorageType enum
        else:
            self.state_storage_type = StateStorageType(os.getenv("STATE_STORAGE_TYPE", "redis").lower())

        self.state = {}

    @cached_property
    def state_handler(self):
        """
        Return the initialized state handler of the given type
        :return:
        """
        if self.state_storage_type == StateStorageType.REDIS:
            return RedisStateHandler()

        if self.state_storage_type == StateStorageType.MONGO:
            return MongoStateHandler()

    def load_state(self, state_key: [str, None] = None, auto_migrate: bool = False, raise_warnings: bool = False) -> dict:
        """
        Load the state from mongo or redis
        :param state_key:
        :param auto_migrate: Set to True if auto-migration from one state storage service to current is needed
        :param raise_warnings: Set to True if size limits warnings should be printed
        :return:
        """
        if not state_key:
            state_key = self.get_formatted_state_key(self.asset_id, self.app_key, self.module_key)

        state = self.state_handler.load_state(self.asset_id, state_key)

        # if auto_migrate key is not passed in, attempt to get from environment
        if not auto_migrate:
            auto_migrate = os.getenv("AUTO_MIGRATE_STATE", "false").lower() == "true"

        if not state and auto_migrate:
            state = self.state_handler.migrate_state(self.asset_id, state_key)
            self.info(self.asset_id, f"Migrated state storage to {self.state_storage_type.value}.")

        size_object = self.state_handler.check_state_size(self.asset_id, state, state_key, raise_warnings=raise_warnings)

        self.debug(self.asset_id, f"Retrieved state from {self.state_storage_type.value} of size {size_object} kb")
        self.state = state
        return state

    def save_state(self, state_key: [str, None] = None, timestamp: [int, None] = None, raise_warnings: bool = True) -> None:
        """
        Save the state to mongo or redis
        :param state_key:
        :param timestamp:
        :param raise_warnings: Set to True if size limits warnings should be printed
        :return:
        """
        if not state_key:
            state_key = self.get_formatted_state_key(self.asset_id, self.app_key, self.module_key)

        size_object = self.state_handler.check_state_size(self.asset_id, self.state, state_key, raise_warnings=raise_warnings)
        self.state_handler.save_state(self.asset_id, self.state, state_key, timestamp)
        self.debug(self.asset_id, f"Saved state to {self.state_storage_type.value} of size {size_object} kb")

    def delete_states(self, asset_id: int, state_keys: [List[str], str]):
        """
        Delete state for current module
        :param state_keys:
        :param asset_id:
        :return:
        """
        if not isinstance(state_keys, list):
            state_keys = [state_keys]

        self.state_handler.delete_states(asset_id, state_keys)
        self.debug(asset_id, f"Deleted state from {self.state_storage_type.value}.")

    @staticmethod
    def get_formatted_state_key(asset_id: int, app_key: str, module_key: [str, None] = None) -> str:
        """
        Returns the state key in Corva naming format
        :param asset_id:
        :param module_key:
        :param app_key:
        :return:
        """
        state_key = "corva/{0}.{1}".format(asset_id, app_key)

        if module_key:
            return "{0}.{1}".format(state_key, module_key)

        return state_key



from copy import deepcopy
from typing import Union, List

from worker.data.operations import get_data_by_path
from worker.data.serialization import serialization
from worker.wellbore.model.enums import HoleType
from worker.wellbore.model.hole_section import HoleSection
from worker.wellbore.sections_mixin import SectionsMixin

DEPTH_THRESHOLD = 50  # FOOT


@serialization
class Hole(SectionsMixin):
    SERIALIZED_VARIABLES = {
        'sections': list
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sections = kwargs.get('sections') or []

    def add_section(self, section: HoleSection):
        """
        Add a new hole section to the end of the hole.
        :param section:
        :return:
        """
        section.top_depth = 0 if not self else self[-1].bottom_depth
        # bottom depth is the source of truth, so we are updating the length
        section.set_length()

        self.sections.append(section)

    def insert_section(self, section: HoleSection, index: int):
        """
        Insert a new hole section in the given index.
        :param section:
        :param index:
        :return:
        """
        if index is not None:
            if self:
                section.top_depth = self[index - 1].bottom_depth
            section.set_bottom_depth()

            self.sections.insert(index, section)
            self.update_depths(start_index=index + 1)

    def get_bottom_depth(self) -> float:
        if not self:
            return 0
        return self[-1].bottom_depth

    def update_depths(self, start_index: int = 0):
        """
        To update the top and bottom depth of the components
        :param start_index:
        :return:
        """
        if start_index < 1:
            self[0].top_depth = 0
            self[0].set_bottom_depth()
            start_index = 1

        for i in range(start_index, len(self)):
            self[i].top_depth = self[i - 1].bottom_depth
            self[i].set_bottom_depth()

    def get_cased_hole(self):
        """
        Get the cased sections of the hole.
        :return:
        """
        cased_hole = Hole()

        for sec in self:
            if sec.hole_type == HoleType.CASED_HOLE:
                cased_hole.add_section(sec)

        return cased_hole

    def get_casing_bottom_depth(self):
        return self.get_cased_hole().get_bottom_depth()

    def is_casing_exist(self) -> bool:
        """
        Get the cased sections of the hole.
        :return:
        """
        return any(sec.hole_type == HoleType.CASED_HOLE for sec in self)

    def update(self, hole_depth: Union[dict, float]) -> bool:
        """
        Update the last open hole segment of the hole.
        :param hole_depth: a wits dict or actual hole depth
        :return: if any update occurred or not
        """
        if self[-1].hole_type != HoleType.OPEN_HOLE:
            return False

        if isinstance(hole_depth, dict):
            hole_depth = get_data_by_path(hole_depth, "data.hole_depth", float)

        # if the hole depth < casing bottom depth, then ignore
        cased_hole_bottom_depth = self.get_cased_hole().get_bottom_depth()
        if hole_depth < cased_hole_bottom_depth:
            return False

        self[-1].set_bottom_depth(hole_depth)
        self[-1].set_length()

        return True

    def set_casings(self, casings: List[dict]):
        """
        Set the casing by providing the list of the jsons
        :param casings: list of casings from API in json dict format
        :return: None
        """
        # TODO we need to consider using 'data.components' if available to set the casings
        self.sections = []

        if not casings:
            return

        # sorting the casings in the ascending order (the smaller the inner_diameter the earlier it is added)
        casings = sorted(casings, key=lambda csg: csg.get('data', {}).get('inner_diameter'))

        # removing the outer casings at the surface
        surface_index = next(
            (
                idx for idx, csg in enumerate(casings)
                if csg.get('data', {}).get('top_depth', 1000) < DEPTH_THRESHOLD
            ), 0
        )
        casings = casings[0: surface_index + 1]

        # finding if a liner is overlapping outer liners
        # this is common in offshore drilling where they set the casings at the sea floor
        index = 1
        while index < len(casings):
            top_depth_inner = casings[index - 1].get('data', {}).get('top_depth', 1000)
            top_depth_outer = casings[index].get('data', {}).get('top_depth', 1000)
            if top_depth_outer + DEPTH_THRESHOLD > top_depth_inner:
                casings.pop(index)
            else:
                index += 1

        # reversing the order (the larger the inner_diameter the earlier it is)
        casings = reversed(casings)

        for csg in casings:
            hole_section = HoleSection(**csg.get('data', {}), hole_type=HoleType.CASED_HOLE)
            self.add_section(hole_section)

    def update_casings(self, casings: 'Hole'):
        """
        To remove the casings of the current hole and
        replace them with the given one
        :param casings:
        :return:
        """
        casings = casings.get_cased_hole()
        # TODO: check if deepcopy is required
        open_holes = deepcopy([sec for sec in self if sec.hole_type == HoleType.OPEN_HOLE])

        self.sections = []
        self.sections.extend(casings)
        self.sections.extend(open_holes)

        self.update_depths()

    def find_section_at_measured_depth(self, measured_depth: float) -> HoleSection:
        return next((sec for sec in self if sec.top_depth < measured_depth <= sec.bottom_depth), self[0])

    def get_min_inner_diameter(self) -> Union[float, None]:
        if not self:
            return None
        return self[-1].inner_diameter

    def get_last_hole_section(self):
        if self:
            return self[-1]
        return None

    def compute_get_total_volume(self, top_depth: float = None, bottom_depth: float = None) -> float:
        """
        Compute the volume of the hole section
        :return: volume of the hole section in FOOT^3
        """
        if top_depth is None and bottom_depth is None:
            return sum(sec.compute_get_volume() for sec in self)

        top_depth = top_depth or 0
        bottom_depth = bottom_depth or self.get_bottom_depth()

        total_volume = 0

        for index, section in enumerate(self):
            if section.bottom_depth <= top_depth:
                continue
            if section.top_depth >= bottom_depth:
                break

            area = section.compute_get_area()  # in INCH^2
            length = min(bottom_depth, section.bottom_depth) - max(top_depth, section.top_depth)  # in FOOT
            volume = area / 144 * length
            total_volume += volume

        return total_volume



from copy import deepcopy
from typing import Union

from worker.data.operations import get_data_by_path, nanround

from worker.data.serialization import serialization
from worker.wellbore.model.annulus import Annulus
from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.hole import Hole


@serialization
class Wellbore:
    SERIALIZED_VARIABLES = {
        'original_drillstring': Drillstring,
        'original_hole': Hole,
        'bit_depth': float,
        'hole_depth': float,
        'last_hole_depth': float,
        'last_non_decreasing_hole_depth': float,
    }

    def __init__(self, **kwargs):
        super().__init__()

        self.original_drillstring: Drillstring = kwargs.get('original_drillstring') or kwargs.get('drillstring')
        self.actual_drillstring: Drillstring = kwargs.get('actual_drillstring')

        self.original_hole: Hole = kwargs.get('original_hole') or kwargs.get('hole')
        self.actual_hole: Hole = kwargs.get('actual_hole')

        self.actual_annulus: Annulus = kwargs.get('actual_annulus')

        self.bit_depth: float = kwargs.get('bit_depth') or 0.0
        self.hole_depth: float = kwargs.get('hole_depth')

        self.last_hole_depth: float = kwargs.get('last_hole_depth') or self.hole_depth
        self.last_non_decreasing_hole_depth: float = kwargs.get('last_non_decreasing_hole_depth') or self.last_hole_depth

        self.update()

    @property
    def annulus(self):
        return self.actual_annulus

    def update(self, bit_depth: Union[float, dict] = None, hole_depth: float = None) -> None:
        """

        :param bit_depth: a wits dict (that will be used for both bit and hole depths) or actual bit depth
        :param hole_depth: actual hole depth value
        :return:
        """
        if self.original_drillstring is None:
            return

        if isinstance(bit_depth, dict):
            bit_depth = get_data_by_path(bit_depth, "data.bit_depth", float)
            hole_depth = get_data_by_path(bit_depth, "data.hole_depth", float)

        self.bit_depth = bit_depth or self.bit_depth or 0.0

        # a qc for the hole depth: if bit depth > hole depth override it with bit depth
        if hole_depth is None:
            hole_depth = self.hole_depth or self.original_hole.get_bottom_depth()
        hole_depth = max(hole_depth, self.bit_depth)
        self.hole_depth = hole_depth

        self.actual_drillstring = deepcopy(self.original_drillstring)
        self.actual_drillstring.update(self.bit_depth)

        self.last_non_decreasing_hole_depth = max(self.last_non_decreasing_hole_depth, self.hole_depth)

        if self.hole_depth > self.last_hole_depth and self.last_hole_depth <= self.last_non_decreasing_hole_depth <= self.hole_depth:
            self.actual_hole = deepcopy(self.original_hole)

        self.actual_hole = deepcopy(self.original_hole)
        self.actual_hole.update(self.hole_depth)

        self.actual_annulus = Annulus(drillstring=self.actual_drillstring, hole=self.actual_hole)

        self.last_hole_depth = self.hole_depth

    def update_casing(self, casings: Hole) -> None:
        """
        If there is any update in the casings
        :param casings:
        :return:
        """
        self.original_hole.update_casings(casings)
        self.actual_hole.update_casings(casings)
        self.update()

    def trim_after_hole_depth_reduction(self, hole_depth: float) -> None:
        self.last_non_decreasing_hole_depth = hole_depth
        self.original_hole = deepcopy(self.actual_hole)

        self.actual_hole.sections = [sec for sec in self.actual_hole.sections if sec.top_depth <= hole_depth]
        if self.actual_hole:
            self.actual_hole[-1].bottom_depth = hole_depth
            self.actual_hole[-1].set_length()

        self.update(hole_depth, hole_depth)

    def compute_get_drillstring_body_volume_change(self, from_bit_depth: float, to_bit_depth: float) -> float:
        """
        Compute the change in the drillstring body volume between two bit depths
        :param from_bit_depth: start bit depth
        :param to_bit_depth: end bit depth
        :return: volume in FOOT^3
        """
        ds_from = deepcopy(self.original_drillstring)
        ds_from.update(from_bit_depth)
        body_volume_from = ds_from.compute_get_body_volume()

        ds_to = deepcopy(self.original_drillstring)
        ds_to.update(to_bit_depth)
        body_volume_to = ds_to.compute_get_body_volume()

        return body_volume_to - body_volume_from

    def __repr__(self):
        return f"Wellbore:\n" \
               f"===Bit Depth / Hole Depth = {nanround(self.bit_depth)} / {nanround(self.hole_depth)}\n" \
               f"===Drillstring:\n{self.actual_drillstring}\n" \
               f"===Hole:\n{self.actual_hole}\n" \
               f"===Annulus:\n{self.actual_annulus}\n\n" \
               f"===Given Drillstring:\n{self.actual_drillstring}\n"



import os
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from worker import exceptions


class API(object):
    HTTP_METHODS = ('get', 'post', 'patch', 'put', 'delete')
    options = {
        "api_url": "API_ROOT_URL",
        "api_key": "API_KEY",
        "app_name": "APP_NAME"
    }

    def __init__(self, *args, **kwargs):
        self.configure(self.options, kwargs)

    def configure(self, options, values):
        for attribute, environment_key in options.items():
            value = values.pop(attribute, os.getenv(environment_key, None))
            if not value:
                error = "No {0} parameter or {1} environment variable defined".format(attribute, environment_key)
                raise exceptions.Misconfigured(error)

            setattr(self, attribute, value)

    def get(self, *args, **kwargs):
        return self.call("get", *args, **kwargs)

    def post(self, *args, **kwargs):
        return self.call("post", *args, **kwargs)

    def patch(self, *args, **kwargs):
        return self.call("patch", *args, **kwargs)

    def put(self, *args, **kwargs):
        return self.call("put", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.call("delete", *args, **kwargs)

    def call(self, method, path, **kwargs):
        content_type = kwargs.pop("content_type", "application/json")
        headers = {
            "Authorization": "API {0}".format(self.api_key),
            "Content-Type": content_type,
            "X-Corva-App": self.app_name
        }

        method = method.lower()
        if method not in self.HTTP_METHODS:
            raise exceptions.APIError("Invalid HTTP method {0}".format(method))

        retry_count = kwargs.pop("retry_count", 5)
        retry_strategy = Retry(
            total=retry_count,
            status_forcelist=[408, 429, 500, 502, 503, 504],
            method_whitelist=["GET", "POST", "PATCH", "PUT", "DELETE"],
            backoff_factor=0.3,
            raise_on_status=False
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http_method = getattr(http, method)

        data = kwargs.pop("data", None)
        if not path.startswith(self.api_url):
            path = "{0}{1}".format(self.api_url, path)

        response = http_method(url=path, data=data, params=kwargs, headers=headers)

        # closing the http connection after a request
        http.close()

        asset_id = kwargs.get("asset_id", "Unknown") or "Unknown"

        if response.status_code == 401:
            raise exceptions.APIError("Unable to reach Corva API")

        if response.status_code == 403:
            raise exceptions.Forbidden("No access to asset {0}".format(asset_id))

        if response.status_code == 404:
            raise exceptions.NotFound("Not Found")

        if not response.ok:
            raise exceptions.APIError(f"{response.status_code} - {response.reason}")

        result = Result(response, **kwargs)
        return result

    def get_by_id(self, path, **kwargs):
        """
        Get a document with the given kwargs ('collection', 'id')
        :param path: API path
        :param kwargs: 'collection' and 'id' kwargs are required
        :return: a result object
        """
        collection = kwargs.pop("collection", None)
        component_id = kwargs.pop("id", None)
        path = "{0}{1}{2}/{3}".format(self.api_url, path, collection, component_id)

        result = self.get(path)
        return result


class Result(object):

    def __init__(self, response, **kwargs):
        self.response = response
        self.params = kwargs
        self.data = None

        try:
            self.data = response.json()
        except Exception:
            raise exceptions.APIError("Invalid API response")

    def __repr__(self):
        return repr(self.data)

    def __iter__(self):
        return iter(self.data)

    @property
    def status(self):
        return self.response.status_code

    @property
    def count(self):
        if not self.data:
            return 0

        if isinstance(self.data, list):
            return len(self.data)

        if isinstance(self.data, dict):
            return 1

        return 0



import json
import os
from typing import List, Tuple, Union

import numpy as np

from worker.data.api import API
from worker.data.enums import EventTypes, Environment
from worker.exceptions import NotFound


def gather_data_for_period(asset_id: int, start: int, end: int, limit: int = 1800, collection: str = "wits") -> list:
    """
    Get the wits data from API for an asset over an interval
    :param asset_id: asset id
    :param start: start timestamp
    :param end: end timestamp
    :param limit: count of the data
    :param collection: any collection
    :return: a list of wits data
    """
    if start >= end:
        return []

    query = '{timestamp#gte#%s}AND{timestamp#lte#%s}' % (start, end)
    worker = API()
    wits_dataset = worker.get(
        path="/v1/data/corva", collection=collection, asset_id=asset_id, sort="{timestamp: 1}", limit=limit, query=query
    ).data

    if not wits_dataset:
        return []

    return wits_dataset


def get_one_data_record(asset_id: int, timestamp_sort: int = -1, collection: str = "wits") -> dict:
    """
    Get the first or last wits record of a given asset
    :param asset_id:
    :param timestamp_sort:
    :param collection:
    :return:
    """
    api_worker = API()
    data = api_worker.get(path='/v1/data/corva/', collection=collection, asset_id=asset_id,
                          sort="{timestamp:%s}" % timestamp_sort, limit=1).data
    if not data:
        return {}

    return data[0]


def delete_collection_data_of_asset_id(asset_id: int, collections: Union[str, list]):
    """
    Delete all the data of a collection for an asset id.
    :param asset_id:
    :param collections: a collection or a list of collections
    :return:
    """
    worker = API()

    if isinstance(collections, str):
        collections = [collections]

    for collection in collections:
        path = "/v1/data/corva/%s" % collection
        query = "{asset_id#eq#%s}" % asset_id
        worker.delete(path=path, query=query)


def point_main_envs(env: str):
    """
    The purpose of this function is to point main environment variables to
    the provided environment. This method updates the following environment
    variables:
    - API_ROOT_URL
    - API_KEY
    - CACHE_URL

    :param env: the environment to point to
    :return:
    """
    if not env:
        return

    # validating the environment
    Environment(env)

    api_url = os.getenv(f"API_ROOT_URL_{env}")
    api_key = os.getenv(f"API_KEY_{env}")
    cache_url = os.getenv(f"CACHE_URL_{env}")

    if not all([api_url, api_key, cache_url]):
        raise ValueError("Missing environment variables!")

    new_envs = {
        'API_ROOT_URL': api_url,
        'API_KEY': api_key,
        'CACHE_URL': cache_url
    }

    os.environ.update(new_envs)


def setup_api_worker(env: str, app_name: str):
    """
    Set up the Corva API worker
    :param env: environment ['local', 'qa', 'staging', 'production']
    :param app_name:
    :return: api worker
    """
    # validating the environment
    Environment(env)

    api_url = os.getenv(f"API_ROOT_URL_{env}")
    api_key = os.getenv(f"API_KEY_{env}")

    options = {
        "api_url": api_url,
        "api_key": api_key,
        "app_name": app_name
    }
    api_worker = API(**options)

    return api_worker


def setup_redis_worker(env: str) -> Tuple:
    """
    Set up the Redis worker
    :param env: environment ['local', 'qa', 'staging', 'production']
    :return: redis worker
    """
    # validating the environment
    Environment(env)

    cache_url = os.getenv(f"CACHE_URL_{env}")

    import redis
    redis_worker = redis.Redis.from_url(cache_url, decode_responses=True)

    return redis_worker


def get_drillstring_by_id(drillstring_id: str) -> Union[dict, None]:
    """
    Get the drillstring from API by providing mongodb _id
    :param drillstring_id: mongodb _id of the drillstring
    :return:
    """
    drillstring = None
    try:
        drillstring = API().get_by_id(
            path="/v1/data/corva/", collection='data.drillstring', id=drillstring_id
        ).data
    except NotFound:
        pass

    return drillstring


def is_number(data):
    """
    Check and return True if data is a number, else return False
    :param data: Input can be string, number or nan
    :return: True or False
    """
    try:
        data_cast = float(data)
        if data_cast >= 0 or data_cast <= 0:  # to make sure it is a valid number
            return True

        return False
    except ValueError:
        return False
    except TypeError:
        return False


def is_finite(data):
    """
    Check if the given data is a finite number
    Note that the string representation of a number is not finite
    :param data:
    :return: True or False
    """
    try:
        return is_number(data) and np.isfinite(data)
    except (TypeError, ValueError):
        return False


def is_int(s: str) -> bool:
    """
    To check if the given string is an integer or not
    :param s:
    :return:
    """
    try:
        int(s)
        return True
    except ValueError:
        return False


def to_number(data):
    """
    Check and return if the data can be cast to a number, else return None
    :param data: Input can be string, number or nan
    :return: A numbers
    """
    if is_number(data):
        return float(data)

    return None


def none_to_nan(data):
    """
    If data is a list, return list with None replaced with nan.
    If data is None, return nan
    :param data:
    :return:
    """
    if isinstance(data, list):
        return [np.nan if e is None else e for e in data]

    if data is None:
        return np.nan

    return data


def get_data_by_path(data: dict, path: str, func=lambda x: x, **kwargs):
    """
    To find the path to a key in a nested dictionary.
    Note that none of the keys should end up in a list
    :param data:
    :param path: path to the final key; example of the paths are:
        'data.X.Y'
        'data.bit_depth'
    :param func: the type of the data (int, str, float, ...)
    :param kwargs: pass default value in case the path not found;
    note that None is an acceptable default
    :return:
    """
    has_default = 'default' in kwargs
    default = kwargs.pop('default', None)

    if not path:
        if has_default:
            return default

        raise KeyError("No key provided")

    keys = path.split('.')

    while keys:
        current_key = keys.pop(0)

        if current_key not in data:
            if has_default:
                return default

            raise KeyError("{0} not found in path".format(current_key))

        data = data.get(current_key)

    if data is None:
        return None

    return func(data)


def is_in_and_not_none(d: dict, key: str):
    """
    An structured way of getting data from a dict.
    :param d: the dictionary
    :param key:
    :return: True or False
    """
    if key in d.keys() and d[key] is not None:
        return True

    return False


def nanround(value, decimal_places=2):
    """
    Similar to python built-in round but considering None values as well
    :param value:
    :param decimal_places:
    :return:
    """
    if is_number(value):
        return round(value, decimal_places)

    return None


def merge_dicts(d1: dict, d2: dict) -> dict:
    """
    Merge two dictionaries
    Note: the 2nd item (d2) has a higher priority to write items with similar keys
    :param d1:
    :param d2:
    :return:
    """
    d = {**d1, **d2}
    return d


def equal(obj1: object, obj2: object, params: List[str]) -> bool:
    """
    To check if two objects are equal by comparing the given parameters.
    :param obj1:
    :param obj2:
    :param params:
    :return:
    """
    if type(obj1) is not type(obj2):
        return False

    return all(getattr(obj1, param) == getattr(obj2, param) for param in params)


def get_cleaned_event_and_type(event) -> Tuple[Union[list, dict], EventTypes]:
    """
    validate and flatten the events and organize the data into a desired format

    Task and generic events format is : dict => {}
    Scheduler events format is: list of list of dict => [[{}]]
    Kafka events format is: list of dict => [{}]
    The above formats can be used to determine the format

    :param event: a scheduler of kafka stream
    :return: event and event_type
    """

    if not event:
        raise ValueError("Empty events")

    if isinstance(event, (str, bytes, bytearray)):
        event = json.loads(event)

    if isinstance(event, dict):
        if "asset_id" in event:
            return event, EventTypes.GENERIC

        if "task_id" in event:
            return event, EventTypes.TASK

        raise TypeError("Missing task_id or asset_id keys in event")

    if not isinstance(event, list):
        raise TypeError("Event is not a list or a dict")

    first_event = event[0]
    if isinstance(first_event, list):
        if first_event[0] and "schedule_start" in first_event[0]:
            return event, EventTypes.SCHEDULER

        raise Exception("Missing scheduler_start key in scheduler event")

    elif isinstance(first_event, dict):
        # new kafka stream format: list of json objects, each with metadata and records
        # event = [{"metadata": { ... }, "records": [ ... ]}, {"metadata": { ... }, "records": [ ... ]}]
        return event, EventTypes.STREAM

    raise TypeError("Event is not either a scheduler or kafka consumer")



import json
import numpy


class JsonEncoder(json.JSONEncoder):
    """
    This encoder can be used to convert incompatible data types to types compatible with json.dumps()
    Use like json.dumps(output, cls=JsonEncoder)
    """
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



class SectionsMixin(object):
    def __init__(self, *args, **kwargs):
        self.sections = []

    def __len__(self):
        return len(self.sections)

    def __getitem__(self, index):
        return self.sections[index]

    def __delitem__(self, index):
        del self.sections[index]

    def __bool__(self):
        return len(self) != 0

    def __repr__(self):
        return "\n".join(str(sec) for sec in self.sections)



from typing import List, Tuple

from worker.data.serialization import serialization
from worker.wellbore.model.ann import Ann
from worker.wellbore.model.drillstring import Drillstring
from worker.wellbore.model.enums import PipeType
from worker.wellbore.model.hole import Hole
from worker.wellbore.measured_depth_finder import get_unique_measured_depths
from worker.wellbore.sections_mixin import SectionsMixin


@serialization
class Annulus(SectionsMixin):
    SERIALIZED_VARIABLES = {
        'sections': list
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sections: List[Ann] = kwargs.get('sections') or []

        drillstring: Drillstring = kwargs.get('drillstring')
        hole: Hole = kwargs.get('hole')
        if drillstring and hole:
            self.create_annulus(drillstring, hole)

    def add_section(self, section: Ann):
        """
        Add a new ann section to the end of the sections.
        :param section:
        :return:
        """
        section.top_depth = 0 if not self else self[-1].bottom_depth
        section.set_bottom_depth()

        self.sections.append(section)

    def insert_section(self, section: Ann, index: int):
        """
        Insert a new pipe element in the given index.
        :param section:
        :param index:
        :return:
        """
        if index is not None:
            if self:
                section.top_depth = self[index - 1].bottom_depth
            section.set_bottom_depth()

            self.sections.insert(index, section)

            for i in range(index + 1, len(self)):
                self[i].top_depth = self[i - 1].bottom_depth
                self[i].set_bottom_depth()

    def check_for_concatenation(self):
        """
        Concatenate the adjacent section with the same ID and OD and hole type
        """
        for i in range(len(self) - 2, -1, -1):
            if self[i].eq_without_length(self[i + 1]):
                self[i].set_bottom_depth(self[i + 1].bottom_depth)
                self[i].set_length()
                del self[i + 1]

    def create_annulus(self, drillstring: Drillstring, hole: Hole):
        """
        Create annulus for a given drillstring and hole configuration
        :return:
        """
        self.sections = []

        md_list = get_unique_measured_depths(drillstring, hole)
        md_list = [md for md in md_list if md <= drillstring[-1].bottom_depth]
        epsilon = 0.0001
        for i in range(len(md_list) - 2):
            drillstring_section = drillstring.find_section_at_measured_depth(md_list[i] + epsilon)
            annulus_section = Ann(**{"top_depth": md_list[i], "bottom_depth": md_list[i + 1]})
            if drillstring_section.pipe_type == PipeType.BIT:
                break
            annulus_section.set_properties_from_pipe_section(drillstring_section)
            annulus_section.set_properties_from_hole(hole.find_section_at_measured_depth(md_list[i] + epsilon))

            self.add_section(annulus_section)

        if not self:
            return

        # extend the last segment to be equal to the bit depth since the bit has been removed
        bit_depth = drillstring.bit_depth
        self[-1].set_bottom_depth(bit_depth)
        self[-1].set_length()

        self.check_for_concatenation()

    def find_section_at_measured_depth(self, measured_depth: float) -> Tuple[Ann, int]:
        for index, ann in enumerate(self):
            if ann.top_depth < measured_depth <= ann.bottom_depth:
                return ann, index

        return self[0], 0

    @property
    def volume(self):
        return sum(sec.volume for sec in self)

    @property
    def bottom_depth(self):
        if not self:
            return 0
        return self[-1].bottom_depth



from copy import deepcopy
from typing import Union, List

from worker.data.operations import get_data_by_path
from worker.data.serialization import serialization
from worker.wellbore.model.drillstring_components import Pipe, Bit, Agitator, PDM, MWD, RSS
from worker.wellbore.model.hole import Hole
from worker.wellbore.sections_mixin import SectionsMixin
from worker.wellbore.model.enums import PipeType


@serialization
class Drillstring(SectionsMixin):
    SERIALIZED_VARIABLES = {
        'sections': list,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.sections: List[Union[Pipe, Bit]] = kwargs.get('sections') or []
        if args:
            self.set_drillstring(args[0])

    def add_section(self, section: Union[Pipe, Bit]):
        """
        Add a new pipe element to the end of the drillstring.
        :param section:
        :return:
        """
        section.top_depth = 0 if not self else self[-1].bottom_depth
        # length is the source of truth, so we are updating the bottom depth
        section.set_bottom_depth()

        self.sections.append(section)

    def insert_section(self, section: Union[Pipe, Bit], index: int):
        """
        Insert a new pipe element in the given index.
        :param section:
        :param index:
        :return:
        """
        if index is not None:
            if self:
                section.top_depth = self[index - 1].bottom_depth
            section.set_bottom_depth()

            self.sections.insert(index, section)

            for i in range(index + 1, len(self)):
                self[i].top_depth = self[i - 1].bottom_depth
                self[i].set_bottom_depth()

    def update(self, bit_depth: Union[dict, float]):
        """
        This method gets the bit depth and chops the current drill string to fit the criteria in the wellbore.
        If the drill string length is more than the bit measured depth it chops the extra length from
        surface. And if it is short in length then it extends the top section to the surface.
        In addition, the indices of the special tools are updated accordingly, when the drillstring is chopped.
        :param bit_depth: a wits dict or actual bit depth
        :return:
        """

        if isinstance(bit_depth, dict):
            bit_depth = get_data_by_path(bit_depth, "data.bit_depth", float)

        self[-1].set_bottom_depth(bit_depth)
        self[-1].set_top_depth()

        for i in range(len(self) - 2, -1, -1):
            self[i].set_bottom_depth(self[i + 1].top_depth)
            self[i].set_top_depth()

            if self[i].bottom_depth < 0:
                del self.sections[i]

        self[0].top_depth = 0
        self[0].set_length()

    @property
    def bit_depth(self):
        if not self:
            return None
        return self[-1].bottom_depth

    def set_drillstring(self, drillstring: dict):
        """
        This gets the whole drillstring json and sets it
        :return:
        """
        if not drillstring:
            return

        self.sections = []

        # family to class mapping
        mapping = {
            'bit': Bit,
            'agitator': Agitator,
            'pdm': PDM,
            'mwd': MWD,
            'rss': RSS
        }

        drillstring_components = drillstring.get("data", {}).get("components", [])
        for component in drillstring_components:
            family = component.get("family", None)
            if family in mapping:
                component = mapping[family](**component)
            else:
                component = Pipe(**component)

            self.add_section(component)

    def find_section_at_measured_depth(self, measured_depth: float) -> Union[Pipe, None]:
        """
        This method returns the pipe at the given measured depth.
        :return:
        """
        if measured_depth > self[-1].bottom_depth:
            return None
        return next((sec for sec in self if sec.top_depth < measured_depth <= sec.bottom_depth), self[0])

    def perform_qc(self, hole: Hole):
        """
        Removes the items that the pipe OD is equal to or greater than the bit/hole size or OD is equal to 0.
        If the pipe id is zero then set it equal to 0.5 of the OD
        :return:
        """
        bit_diameter = self.get_bit().size

        self.sections = [
            sec for sec in self.sections
            if 0 < sec.outer_diameter < bit_diameter
        ]

        for component in self.sections:
            if component.inner_diameter == 0:
                component.inner_diameter = component.outer_diameter / 2

    def compute_get_inside_volume(self):
        """
        Compute and get the inside volume of the pipes of the drill string only,
        not the body and outside of it.
        :return: volume in FOOT^3
        """
        return sum(sec.compute_get_volume() for sec in self)

    def compute_get_body_volume(self):
        """
        Compute and get the body volume of the pipes of the drill string only,
        not inside and outside of it.
        :return: volume in FOOT^3
        """
        ds = self.get_a_copy_without_bit()
        return sum(sec.compute_body_volume_tj_adjusted() for sec in ds)

    def get_a_copy_without_bit(self):
        """
        To get a deep copy of the current drillstring without the bit
        :return:
        """
        ds = deepcopy(self)
        bit = ds.get_bit()
        if bit:
            ds.sections.remove(bit)
        return ds

    def get_bit(self):
        """
        Get the bit element from the drillstring
        :return:
        """
        if not self:
            return None

        bit = self[-1]
        if isinstance(bit, Bit):
            return bit

        return None

    def get_bha_length(self) -> float:
        """
        To get the length of the BHA
        :return:
        """
        dp_indices = [index for index, pipe in enumerate(self) if pipe.pipe_type == PipeType.DP]
        last_dp_index = -1
        if dp_indices:
            last_dp_index = max(dp_indices)
        length = sum(el.length for el in self[last_dp_index + 1:])
        return length



import os
import datetime

from worker.data.operations import is_number


class LoggingMixin(object):
    LOGGING_LEVELS = [
        "all",  # Show all messages
        "debug",  # Show debug, info, warnings, errors, and fatal messages (Equal to all)
        "info",  # Show info, warnings, errors, and fatal messages
        "warn",  # Show warnings, errors, and fatal messages
        "error",  # Show errors and fatal messages
        "fatal",  # Show only fatal messages
        "off"  # No messages
    ]

    def __init__(self, *args, **kwargs):
        self.logging_level = os.getenv("LOGGING_LEVEL", "info").lower()
        self.logging_asset_id = os.getenv("LOGGING_ASSET_ID", 0)

        if self.logging_level not in self.LOGGING_LEVELS:
            self.logging_level = "off"

        if is_number(self.logging_asset_id):
            self.logging_asset_id = int(self.logging_asset_id)

        super().__init__(*args, **kwargs)

    def all(self, asset_id, text):
        self.call('all', asset_id, text)

    def debug(self, asset_id, text):
        self.call('debug', asset_id, text)

    def info(self, asset_id, text):
        self.call('info', asset_id, text)

    def warn(self, asset_id, text):
        self.call('warn', asset_id, text)

    def error(self, asset_id, text):
        self.call('error', asset_id, text)

    def fatal(self, asset_id, text):
        self.call('fatal', asset_id, text)

    def off(self, asset_id, text):
        pass

    def call(self, method, asset_id, text):
        if method == 'off':
            return

        method_check = str(method) in self.LOGGING_LEVELS[self.LOGGING_LEVELS.index(self.logging_level):]
        if not method_check:
            return

        asset_id_check = (asset_id == self.logging_asset_id) or (self.logging_asset_id == 'all')
        if not asset_id_check:
            return

        self.log(asset_id, text)

    @staticmethod
    def log(asset_id, text):
        print(datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%f"), 'asset_id_{} -> {}'.format(asset_id, text))



import json

from worker.state.mixins import StateMixin


def file_to_json(file_name):
    with open(file_name, mode='r') as file:
        _json = json.load(file)
        return _json


def get_last_processed_timestamp(asset_id: int, state_key: str, state_storage_type=None):
    """
    Get the last_processed_timestamp from state storage
    :param asset_id:
    :param state_key:
    :param state_storage_type:
    """
    try:
        state_app = StateMixin(state_storage_type)
        state_app.asset_id = asset_id
        previous_state = state_app.load_state(state_key=state_key, auto_migrate=False)
        return previous_state.get('last_processed_timestamp', None)
    except Exception as ex:
        print(f"Error occurred while reading state from {state_storage_type}!\nError: {ex}")

    return None


def create_scheduler_events(asset_id, start_timestamp, end_timestamp, step):
    """
    Creating scheduler events
    :param asset_id:
    :param start_timestamp:
    :param end_timestamp:
    :param step:
    :return:
    """
    if start_timestamp > end_timestamp:
        raise ValueError(f"start_timestamp ({start_timestamp}) is greater than end_timestamp ({end_timestamp})!")
    if step <= 0 or step > 3600:
        raise ValueError(f"step ({step}) is outside the (0, 3600] range.")

    triggers = range(start_timestamp, end_timestamp, step)
    events = [[[{"asset_id": asset_id, "schedule_start": 1000 * trigger, "schedule_end": 1000 * (trigger + step)}]]
              for trigger in triggers]
    return events



class RollbarMixin(object):
    def __init__(self, *args, **kwargs):
        self.rollbar = kwargs.pop('rollbar', None)
        super().__init__(*args, **kwargs)

    def is_rollbar(self) -> bool:
        """
        To check if rollbar is available or not
        :return: if rollbar is available
        """
        return self.rollbar and self.rollbar.SETTINGS.get('enabled')

    def track_message(self, message: str, level: str):
        """
        To send a message to rollbar
        :param message:
        :param level: any of the following levels:
        ['critical', 'error', 'warning', 'info', 'debug', 'ignored']
        :return:
        """
        # Levels:
        if not self.is_rollbar():
            print(f"{level} - {message}")
            return

        level = level.lower()

        self.rollbar.report_message(message, level)

    def track_error(self, message: str = None):
        if not self.is_rollbar():
            raise

        self.rollbar.report_exc_info(extra_data=message, level='error')



import os
import time

from worker import Module
from worker.app.modules.trigger import Trigger
from worker.state.enums import StateStorageType


class TestMongo(Module):
    module_key = "mongo-tester"
    collection = "None"

    module_state_fields = {
        'message': str,
    }

    # override
    def run_module(self, event, beginning_results):
        self.state_storage_type = StateStorageType.MONGO
        start_timer = time.time()
        reset_states(self)
        test_save_load(self)
        test_delete_states(self)
        self.state = {"message": f"This state is auto saved by module.run {self.module_key}"}
        print(f"Mongo tests completed in {int(time.time() - start_timer)} seconds.\n")

    def run(self, wits_stream: list):
        self.run_module(wits_stream, [])


class TestRedis(Module):
    module_key = "redis-tester"
    collection = "None"

    module_state_fields = {
        'message': str,
    }
    os.environ["STATE_STORAGE_TYPE"] = StateStorageType.REDIS.value

    # override
    def run_module(self, event, beginning_results):
        self.state_storage_type = StateStorageType.REDIS
        start_timer = time.time()
        reset_states(self)
        test_save_load(self)
        test_delete_states(self)
        self.state = {"message": f"This state is auto saved by module.run {self.module_key}"}
        print(f"Redis tests completed in {int(time.time() - start_timer)} seconds.\n")

    def run(self, wits_stream: list):
        self.run_module(wits_stream, [])


class TestMigration(Trigger):
    module_key = "migration-tester"
    collection = "None"

    module_state_fields = {
        'message': str,
    }

    # override
    def run_module(self, event, beginning_results):
        start_timer = time.time()
        global_state = {
            "asset_id": event.get("asset_id")
        }

        # Migration Tests
        print("Test auto migration of state")
        asset_id = global_state["asset_id"]
        module_key = self.module_key

        # Initialize and reset redis app
        redis_module = TestRedis(global_state, state_storage_type=StateStorageType.REDIS.value)
        redis_module.module_key = module_key
        redis_state_key = redis_module.get_formatted_state_key(global_state.get("asset_id"), redis_module.app_key, redis_module.module_key)
        redis_module.delete_states(asset_id, redis_state_key)
        redis_module.load_state(auto_migrate=False)

        # Initialize and reset mongo app
        mongo_module = TestMongo(global_state, state_storage_type=StateStorageType.MONGO.value)
        mongo_module.module_key = module_key
        mongo_state_key = mongo_module.get_formatted_state_key(global_state.get("asset_id"), mongo_module.app_key, mongo_module.module_key)
        mongo_module.delete_states(asset_id, mongo_state_key)
        mongo_module.load_state(auto_migrate=False)
        mongo_state = {"message": "This state is on mongo"}

        # Migration from mongo to redis
        # create a state on mongo
        # delete the state from redis
        # try to load from redis, it should automigrate from mongo to redis
        print("Test auto-migration from mongo to redis")
        mongo_module.state = mongo_state
        mongo_module.save_state()
        redis_module.delete_states(asset_id, redis_state_key)
        state = redis_module.load_state(auto_migrate=True)
        assert mongo_state == state

        # Initialize and reset redis app
        redis_module = TestRedis(global_state, state_storage_type=StateStorageType.REDIS.value)
        redis_module.module_key = module_key
        redis_module.delete_states(asset_id, redis_state_key)
        redis_module.load_state(auto_migrate=False)
        redis_state = {"message": "This state is on redis"}

        # Initialize and reset mongo app
        mongo_module = TestMongo(global_state, state_storage_type=StateStorageType.MONGO.value)
        mongo_module.module_key = module_key
        mongo_module.delete_states(asset_id, mongo_state_key)
        mongo_module.load_state(auto_migrate=False)

        # Migration from redis to mongo
        # create a state on redis
        # delete the state from mongo
        # try to load from mongo, it should automigrate from redis to mongo
        print("Test auto-migration from redis to mongo")
        redis_module.state = redis_state
        redis_module.save_state()
        mongo_module.delete_states(asset_id, mongo_state_key)
        state = mongo_module.load_state(auto_migrate=True)
        assert redis_state == state

        self.state = {"message": f"This state is auto saved by module.run {self.module_key}"}
        print(f"Migration tests completed in {int(time.time() - start_timer)} seconds.\n")


def test_delete_states(main_module):
    # Deletion Tests
    state_storage_type = main_module.state_storage_type.value
    print(f"Delete test for {state_storage_type}")

    asset_id = main_module.global_state["asset_id"]

    modules = []
    for i in range(5):
        _module_key = f"delete.test{i}"
        module = main_module.__class__(main_module.global_state, state_storage_type=state_storage_type)
        module.module_key = _module_key
        state_key = module.get_formatted_state_key(asset_id, module.app_key, _module_key)
        state = {"message": f"{module.state_storage_type.value} state {state_key}"}
        module.state = state
        module.save_state()
        modules.append(module)
        time.sleep(0.1)  # back-off to prevent blocking by redis/mongo

    # Delete Multiple
    state_keys = [
        main_module.get_formatted_state_key(asset_id, module.app_key, module.module_key)
        for module in modules
    ]
    multiple_state_keys = state_keys[0:4]

    main_module.delete_states(asset_id, multiple_state_keys)
    for i in range(4):
        _module_key = f"delete.test{i}"
        current_state = modules[i].load_state(auto_migrate=False)
        assert current_state == {"message": None}

    # Delete Single
    main_module.delete_states(asset_id, state_keys[4])
    current_state = modules[4].load_state(auto_migrate=False)
    assert current_state == {"message": None}


def reset_states(module):
    # Initialize and reset states
    print(f"Initialize test for {module.state_storage_type.value}")
    asset_id = module.global_state["asset_id"]
    state_key = module.get_formatted_state_key(asset_id, module.app_key, module.module_key)
    module.delete_states(asset_id, state_key)
    module.load_state()


def test_save_load(module):
    # Save/load tests
    print(f"Save/load test for {module.state_storage_type.value}")
    current_state = {"message": f"This state is on {module.state_storage_type.value}"}
    module.state = current_state
    module.save_state()
    state = module.load_state(auto_migrate=False)
    assert current_state == state



import json
import os
from typing import List

from cached_property import cached_property

from worker import API
from worker.state.enums import StateStorageType
from worker.state.handlers.__init__ import StateHandler


class MongoStateHandler(StateHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.state_storage_type = StateStorageType.MONGO
        state_storage_limit_fatal = os.getenv("STATE_STORAGE_CRITICAL_LIMIT", 100_000.0)  # 100 MB
        state_storage_limit_warning = os.getenv("STATE_STORAGE_WARNING_LIMIT", 10_000.0)  # 10 MB
        self.set_size_limit("fatal", float(state_storage_limit_fatal))
        self.set_size_limit("warning", float(state_storage_limit_warning))

        self.api_path = "/v1/data/corva/temporary-state"

        self.created_timestamp = None
        self.mongo_id = None

    @cached_property
    def api(self):
        """
        Get the mongo/api connection setup
        :return:
        """
        return API()

    def load_state(self, asset_id: int, state_key: str) -> dict:
        """
        Load the state from mongo
        :param asset_id:
        :param state_key:
        :return:
        """
        query = "{state_key#eq#'%s'}" % state_key
        asset_id = asset_id
        sort = "{timestamp: -1}"
        state = self.api.get(path=self.api_path, asset_id=asset_id, sort=sort, query=query, limit=1).data

        if not state:
            return {}

        state = state[-1]
        self.mongo_id = state.get("_id")
        self.created_timestamp = state.get("timestamp")

        return state.get("data", {}).get("state", {})

    def save_state(self, asset_id: int, state: dict, state_key: str, timestamp: int):
        """
        Save the state to mongo
        :param asset_id:
        :param state:
        :param state_key:
        :param timestamp:
        :return:
        """
        state = {
            "state": state
        }
        record = {
            "collection": "temporary-state",
            "state_key": state_key,
            "updated_timestamp": timestamp,
            "version": 1,
            "provider": "corva",
            "asset_id": asset_id,
            "data": state,
        }

        if self.mongo_id:
            record["_id"] = self.mongo_id

        # if a blank state was loaded set the created timestamp to current timestamp
        record["timestamp"] = self.created_timestamp or timestamp
        response = self.api.post(path="/v1/data/corva", data=json.dumps(record)).data

        self.mongo_id = response.get("_id")
        return response

    def migrate_state(self, asset_id: int, state_key: str, from_state_storage=StateStorageType.REDIS) -> dict:
        """
        Attempt to get state from redis and save it to mongo
        This is needed for assets that already running, and saving state to redis
        Useful if the lambda has been setup for mongo but previous live assets were saving to redis
        :param asset_id:
        :param state_key:
        :param from_state_storage:
        :return:
        """
        if from_state_storage != StateStorageType.REDIS:
            return {}

        from worker.state.handlers.redis import RedisStateHandler
        state_handler = RedisStateHandler()

        state = state_handler.load_state(asset_id, state_key)

        created_timestamp = None
        if isinstance(state, dict):
            created_timestamp = state.get("timestamp") or state.get("last_processed_timestamp")

        if not state:
            state = {}

        self.save_state(asset_id, state, state_key, created_timestamp)
        return state

    def delete_states(self, asset_id: int, state_keys: List[str]):
        """
        Delete given state keys from mongo
        :param asset_id:
        :param state_keys:
        :return:
        """
        delete_count = 0
        for state_key in state_keys:
            delete_count += self._delete_state(asset_id, state_key)

        self.debug(asset_id, f"Deleted {delete_count} states")
        return delete_count

    def _delete_state(self, asset_id: int, state_key: str) -> int:
        """
        Delete state from mongo
        :param asset_id:
        :param state_key:
        :return:
        """
        query = "{asset_id#eq#%s}AND{state_key#eq#'%s'}" % (asset_id, state_key)
        response = self.api.delete(path=self.api_path, query=query)
        delete_count = response.data.get("deleted_count", 0)
        return delete_count



"""
The purpose of this simple test app is to compute Surface MSE
every 30 seconds of wits stream records. Note that this parameter
is only generated when the state is Drilling (either Rotary or Slide).
"""
from testcase.app.app_constants import constants  # noqa: F401



import json
import sys
from typing import List

from worker.mixins.logging import LoggingMixin
from worker.mixins.rollbar import RollbarMixin


class StateHandler(RollbarMixin, LoggingMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.size_limits = {}
        self.state_storage_type = None

    def load_state(self, asset_id: int, state_key: str) -> dict:
        raise NotImplementedError("load_state method must be implemented")

    def save_state(self, asset_id: int, state: dict, timestamp: [int, None], state_key: str) -> None:
        raise NotImplementedError("save_state method must be implemented")

    def check_state_size(self, asset_id, state, state_key, raise_warnings=True):
        """
        Check the size of the state dictionary and generate warnings if necessary
        :param asset_id:
        :param state:
        :param state_key:
        :param raise_warnings:
        :return:
        """
        size_object = sys.getsizeof(json.dumps(state)) / 1024

        if not raise_warnings:
            return size_object

        size_limit = self.size_limits["fatal"]
        if size_object > size_limit:
            message = f"State to {self.state_storage_type.value} of state_key {state_key} is of size {size_object} kb > {size_limit} kb."
            self.fatal(asset_id, message)
            self.track_message(message, level="critical")
            return size_object

        size_limit = self.size_limits["warning"]
        if size_object > size_limit:
            message = f"State to {self.state_storage_type.value} of state_key {state_key} is of size {size_object} kb > {size_limit} kb."
            self.warn(asset_id, message)
            return size_object

        return size_object

    def migrate_state(self, asset_id: int, state_key: str, from_state_storage) -> dict:
        raise NotImplementedError("migrate_state method is not implemented")

    def delete_states(self, asset_id: int, state_keys: List[str]):
        raise NotImplementedError("delete_states method is not implemented")

    def set_size_limit(self, limit_type: str, state_storage_limit: float):
        self.size_limits[limit_type] = state_storage_limit



from copy import deepcopy

import numpy as np

from worker.data.serialization import serialization
from worker.data.unit_conversions import US_LIQUID_GAL_to_INCH3
from worker.wellbore.model.element import Element
from worker.wellbore.model.enums import PipeType, PipeMaterial
from worker.data.operations import nanround, equal


@serialization
class Pipe(Element):
    SERIALIZED_VARIABLES = {
        'top_depth': float,
        'bottom_depth': float,
        'length': float,

        'inner_diameter': float,
        'outer_diameter': float,

        'joint_length': float,
        'max_outer_diameter': float,
        'inner_diameter_tooljoint': float,
        'outer_diameter_tooljoint': float,
        'tool_joint_length_per_joint': float,
        'linear_mass_in_air': float,
        'order': float,

        'pipe_material': PipeMaterial,
        'grade': float,

        'tool_joint_length_ratio': float,
        'pipe_type': PipeType,
    }

    """
    The pipe object can be used for the following components directly:
    - drill pipe (DP)
    - drill collar (DC)
    - heavy weight drill pipe (HWDP)
    - subs
    - stabilizers

    This class is extended by the other drillpipe components.
    """
    def __init__(self, **kwargs):
        super(Pipe, self).__init__(**kwargs)

        self.inner_diameter = kwargs['inner_diameter']
        self.outer_diameter = kwargs['outer_diameter']

        self.joint_length = kwargs.get('component_length', kwargs.get('joint_length'))
        self.max_outer_diameter = kwargs.get('max_outer_diameter')
        self.inner_diameter_tooljoint = kwargs.get('inner_diameter_tooljoint')
        self.outer_diameter_tooljoint = kwargs.get('outer_diameter_tooljoint')
        self.tool_joint_length_per_joint = kwargs.get('length_tooljoint') or kwargs.get('tool_joint_length_per_joint')
        self.linear_mass_in_air = kwargs.get('linear_weight', kwargs.get('linear_mass_in_air'))
        self.order = kwargs.get('order')

        self.pipe_material = kwargs.get('material')
        self.grade = kwargs.get('grade')

        self.set_tool_joint_length_ratio()

        family = kwargs.get('family') or ""
        self.pipe_type = PipeType.determine_type(family)

    def get_max_outer_diameter(self):
        return self.max_outer_diameter or self.outer_diameter_tooljoint or self.outer_diameter

    def set_tool_joint_length_ratio(self):
        if self.tool_joint_length_per_joint is None or not self.joint_length:
            self.tool_joint_length_ratio = 0
        else:
            self.tool_joint_length_ratio = self.tool_joint_length_per_joint / self.joint_length

    def is_valid_tool_joint(self):
        """
        If there is a valid tool joint or not
        :return:
        """
        if self.outer_diameter_tooljoint is None:
            return False
        if self.inner_diameter_tooljoint is None:
            return False
        if not self.tool_joint_length_ratio:
            return False

        return True

    def get_pipe_id_area(self):
        """
        Compute the total inner area of the pipe body
        :return: inner area of the pipe body in INCH^2
        """
        return np.pi / 4 * (self.inner_diameter ** 2)

    def get_pipe_od_area(self):
        """
        Compute the total outer area of the pipe body
        :return: outer area of the pipe body in INCH^2
        """
        return np.pi / 4 * (self.outer_diameter ** 2)

    def get_pipe_tool_joint(self):
        """
        Create a pipe with its id and od equal to the tool joint values (if available)
        :return:
        """
        p = deepcopy(self)
        p.inner_diameter = p.inner_diameter_tooljoint or p.inner_diameter
        p.outer_diameter = p.outer_diameter_tooljoint or p.outer_diameter
        return p

    def compute_inner_area_tool_joint_adjusted(self):
        """
        Compute the inner area of the pipe adjusted for the tool joint
        :return: the area in INCH^2
        """
        area_body = self.get_pipe_id_area()  # in INCH^2
        if not self.is_valid_tool_joint():
            return area_body

        area_tj = np.pi / 4 * (self.inner_diameter_tooljoint ** 2)  # in INCH^2
        return (
            area_body * (1 - self.tool_joint_length_ratio)
            + area_tj * self.tool_joint_length_ratio
        )

    def compute_outer_area_tool_joint_adjusted(self):
        """
        Compute the outer area of the pipe adjusted for the tool joint
        :return: the area in INCH^2
        """
        area_body = self.get_pipe_od_area()  # in INCH^2
        if not self.is_valid_tool_joint():
            return area_body

        area_tj = np.pi / 4 * (self.outer_diameter_tooljoint ** 2)  # in INCH^2
        r = self.tool_joint_length_ratio
        return area_body * (1 - r) + area_tj * r

    def compute_inner_diameter_tool_joint_adjusted(self):
        """
        Compute the inner diameter of the pipe adjusting for the tool joint
        :return: the adjusted inner diameter of the pipe in INCH
        """
        di_body = self.inner_diameter
        if not self.is_valid_tool_joint():
            return di_body

        di_tj = self.inner_diameter_tooljoint
        r = self.tool_joint_length_ratio
        return di_body * (1 - r) + di_tj * r

    def compute_outer_diameter_tool_joint_adjusted(self):
        """
        Compute the outer diameter of the pipe adjusting for the tool joint
        :return: the adjusted outer diameter of the pipe in INCH
        """
        do_body = self.outer_diameter
        if not self.is_valid_tool_joint():
            return do_body

        do_tj = self.outer_diameter_tooljoint
        r = self.tool_joint_length_ratio
        return do_body * (1 - r) + do_tj * r

    def compute_body_cross_sectional_area_tj_adjusted(self):
        """
        Compute body cross sectional area adjusted for the tool joint
        :return: area in INCH^2
        """
        return self.compute_outer_area_tool_joint_adjusted() - self.compute_inner_area_tool_joint_adjusted()

    # override
    def compute_get_area(self):
        return self.compute_inner_area_tool_joint_adjusted()

    def compute_body_volume_tj_adjusted(self):
        """
        Compute body volume adjusted for the tool joint
        :return: volume in FOOT^3
        """
        return self.compute_body_cross_sectional_area_tj_adjusted() / 144 * self.length

    def compute_outer_volume_tj_adjusted(self):
        """
        Compute outer volume adjusted for the tool joint
        :return: volume in FOOT^3
        """
        return self.compute_outer_area_tool_joint_adjusted() / 144 * self.length

    def compute_inner_volume_tj_adjusted(self):
        """
        Compute inner volume adjusted for the tool joint
        :return: volume in FOOT^3
        """
        return self.compute_inner_area_tool_joint_adjusted() / 144 * self.length

    # override
    def compute_get_volume(self):
        return self.compute_inner_volume_tj_adjusted()

    def calculate_linear_mass(self):
        """
        Compute the pipe linear mass in air from pipe materials
        :return: pipe linear mass in LB/FOOT
        """
        density = self.pipe_material.get_density()  # PPG
        pipe_body_area = self.get_pipe_body_area()  # INCH^2
        linear_mass = density * pipe_body_area * (US_LIQUID_GAL_to_INCH3 * 12)  # lb/ft
        return linear_mass

    def __eq__(self, other):
        if not isinstance(other, (Pipe, Bit)):
            return False

        if self.eq_without_length(other) is False:
            return False

        return self.length == other.length

    def eq_without_length(self, other):
        params = [
            'pipe_type',
            'inner_diameter',
            'outer_diameter',
            'inner_diameter_tooljoint',
            'outer_diameter_tooljoint',
            'linear_mass_in_air'
        ]
        return equal(self, other, params)

    def eq_id_and_od(self, other):
        params = [
            'inner_diameter',
            'outer_diameter'
        ]
        return equal(self, other, params)

    def __repr__(self):
        return (
            f"{self.pipe_type.name:13}"
            f"{nanround(self.inner_diameter, 2):>6} in / "
            f"{nanround(self.outer_diameter, 2):>6} in, "
            f"{nanround(self.linear_mass_in_air, 2):>6} lb/ft, "
            f"{super().__repr__()}"
        )


@serialization
class MWD(Pipe):
    # TODO extra items to be implemented later
    pass


@serialization
class PDM(Pipe):
    # TODO extra items to be implemented later
    pass


@serialization
class RSS(Pipe):
    # TODO extra items to be implemented later
    pass


@serialization
class Agitator(Pipe):
    pass


@serialization
class Bit(Element):
    SERIALIZED_VARIABLES = {
        'top_depth': float,
        'bottom_depth': float,
        'length': float,

        'size': float,
        'tfa': float,
        'pipe_type': PipeType
    }

    # TODO extra items to be implemented later
    def __init__(self, **kwargs):
        super(Bit, self).__init__(**kwargs)

        self.size = float(kwargs['size'])
        self.tfa = float(kwargs['tfa'])
        self.pipe_type = PipeType.BIT

    # override
    def compute_get_area(self):
        return 0

    # override
    def compute_get_volume(self):
        return 0

    def __repr__(self):
        return (
            f"{self.pipe_type.name:13}"
            f"{nanround(self.size, 2)} in - "
            f"tfa={nanround(self.tfa, 2)} in^2, "
            f"{super().__repr__()}"
        )



from typing import List

import numpy as np

from worker import API
from worker.app import TimeActivityModule
from testcase.app.app_constants import constants
from worker.data.operations import nanround
from worker.exceptions import Misconfigured


class SurfaceMSE(TimeActivityModule):
    module_key = "mse"
    collection = "drilling-efficiency.mse"

    module_state_fields = {
        'last_processed_timestamp': int,

        'last_exported_timestamp': int,
        'last_exported_depth': float,

        'active_string_type': str,
        'active_string_id': str,
        'drillstring_number': float,
    }

    # override
    def run_module(self, event, beginning_results):
        self.results = beginning_results

        valid_activities = constants.get(f"drilling-efficiency.{self.module_key}.drilling-activity")
        last_exported_timestamp = self.state.get('last_exported_timestamp', -1)

        valid_records = [
            r
            for r in event
            if r.get('data', {}).get('state', '') in valid_activities and r['timestamp'] > last_exported_timestamp
        ]
        if not valid_records:
            return self.results

        valid_timestamps = [
            r['timestamp']
            for r in valid_records
        ]

        drillstring = get_external_data(wits=valid_records[-1])
        bit_size = extract_bit_size(drillstring)

        updating_frequency_duration = constants.get(f"drilling-efficiency.{self.module_key}.export-duration")

        if last_exported_timestamp <= 0:
            last_exported_timestamp = min(valid_timestamps) - updating_frequency_duration

        max_timestamp = max(valid_timestamps)

        time_check = last_exported_timestamp
        while (time_check + updating_frequency_duration) <= max_timestamp:
            start = time_check  # start of the interval
            end = start + updating_frequency_duration  # end of the interval

            segment_data = [r for r in valid_records if start <= r.get('timestamp', 0) <= end]
            if not segment_data:
                continue

            mse = self.compute_average_mse(segment_data, bit_size)

            segment_last_hole_depth = segment_data[-1].get('data', {}).get('hole_depth')
            output = self.build_empty_output(segment_data[-1])
            output['data'] = {
                'surface_mse': nanround(mse, decimal_places=1),  # in psi
                'depth': segment_last_hole_depth
            }

            self.results.append(output)

            self.set_state('last_exported_timestamp', time_check)
            self.set_state('last_exported_depth', segment_last_hole_depth)
            time_check = end

        self.debug(self.global_state.get('asset_id'), self.results)
        return self.results

    def compute_average_mse(self, wits_records: List[dict], bit_size: float) -> float:
        """
        Compute MSE for a list of drilling wits records
        :param wits_records: WITS records
        :param bit_size: bit size in INCH
        :return: MSE in PSI
        """
        channels = ['weight_on_bit', 'rotary_rpm', 'rotary_torque', 'rop']
        data = {k: [] for k in channels}
        for wits in wits_records:
            wits_data = wits.get('data', {})
            for ch in channels:
                data[ch].append(wits_data.get(ch))
        ave_data = {k: np.nanmean(data[k]) for k in channels}

        mse = compute_mse(bit_size,
                          rop=ave_data['rop'], rpm=ave_data['rotary_rpm'],
                          wob=ave_data['weight_on_bit'], torque=ave_data['rotary_torque'])

        return mse


def get_external_data(wits: dict) -> dict:
    """
    TODO
    :param wits:
    :return:
    """
    drillstring__id = wits.get('metadata', {}).get('drillstring', None)
    if drillstring__id is None:
        raise Misconfigured(f"Wrong drillstring _id ({drillstring__id})!")

    worker = API()
    drillstring = worker.get_by_id(
        path="/v1/data/corva/", **{'collection': 'data.drillstring', 'id': drillstring__id}
    ).data  # dict
    return drillstring


def extract_bit_size(drillstring: dict) -> float:
    """
    Get the bit size from a drillstring.
    :param drillstring: the dictionary holding the drillstring info
    :return: bit size
    """
    components = drillstring.get('data', {}).get('components', [])
    bit = list(filter(lambda x: x['family'] == 'bit', components))
    if not bit:
        raise Misconfigured(f"No bit in the drillstring: {drillstring}")
    return bit[0].get('size', None)


def compute_mse(bit_size, rop, rpm, wob, torque, factor=0.35):
    """
    Compute the MSE
    :param bit_size: bit size in INCH
    :param rop: WITS ROP in FOOT/HOUR
    :param rpm: WITS Rotary RPM in RPM
    :param wob: WITS Weight on Bit in KIPS
    :param torque: WITS Rotary Torque in K-FT-LBF
    :param factor: Based on the recommendation of Dupriest et. al. paper
    (Dupriest, Fred E., and William L. Koederitz. "Maximizing drill rates
     with real-time surveillance of mechanical specific energy."
     SPE/IADC Drilling Conference. Society of Petroleum Engineers, 2005.)
      a factor of <b>0.35</b> is applied for surface MSE.
    :return: MSE in PSI
    """
    bit_area = np.pi / 4 * bit_size ** 2

    if rop < 0:
        return None

    mse = factor * ((wob / bit_area) + (120 * np.pi * rpm * 1000 * torque) / (bit_area * rop))
    return mse



import os
import time
import unittest

from testcase.state_testing_app.app_lambda import lambda_handler


class TestStateMixin(unittest.TestCase):
    def test_state_mixin(self):
        asset_id = 29399830

        os.environ["LOGGING_ASSET_ID"] = str(asset_id)
        os.environ["LOGGING_LEVEL"] = "debug"
        os.environ["STATE_STORAGE_CRITICAL_LIMIT"] = "10"
        event = [{
            "asset_id": asset_id,
            "timestamp": 2
        }]
        start_timer = time.time()

        # loading of constants should happen here to avoid conflicts
        # although this import is unused but it is required to reset the constants
        from testcase.state_testing_app.app_constants import constants  # noqa: F401
        lambda_handler(event, None)
        print(f"Finished testing StateMixin in {int(time.time() - start_timer)} seconds.")



from itertools import groupby

from worker.app.modules.__init__ import Module
from worker.data.api import API
from worker.data import operations
from worker.data.operations import get_drillstring_by_id
from worker.framework import constants


class ActivityModule(Module):
    """
    This is an abstract base module that needs to be extended by an actual module.
    """
    # module_key is used for redis access and state of this module
    module_key = "activity_module"
    collection = "activity_module_collection"

    # override
    def run(self, wits_stream: list):
        """
        :param wits_stream: a wits stream event
        :return:
        """
        super().run(wits_stream)

        dataset = self.load_dataset(wits_stream)

        # this guarantees every group has the same config
        config_grouped_dataset = self.group_data_stream_based_on_config(dataset)

        for group in config_grouped_dataset:
            results = self.check_for_string_change(group)

            if self.should_run_processor(group):
                results = self.run_module(group, results)

            self.save_state()
            self.store_output(self.global_state['asset_id'], results)

    @staticmethod
    def group_data_stream_based_on_config(dataset):
        """Group our datasets by the metadata such as drillstring, and then reflatten to be a list of lists
        for processing"""

        groups = groupby(dataset, lambda x: x['metadata'])
        grouped_dataset = [list(dataset) for group, dataset in groups]

        return grouped_dataset

    # override
    def should_run_processor(self, event):
        running_string = constants.get("{}.{}.running-string".format(self.app_key, self.module_key))
        if running_string and self.state.get('active_string_type', '') in running_string:
            return True

        return False

    def check_for_string_change(self, data):
        """
        To check if the active string in the wellbore has been changed and
        export proper text to the output.
        Note that the last_exported_timestamp is not set here.
        Note that all the configs of this data have the same config properties.
        :param data:
        :return:
        """

        previous_string = self.state.get('active_string_id', "")
        running_string = constants.get("{}.{}.running-string".format(self.app_key, self.module_key))
        reset_config = constants.get("{}.{}.reset-config".format(self.app_key, self.module_key))

        if not running_string:
            return []

        if not data:
            return []

        first_wits = data[0]
        drillstring_id = first_wits.get('metadata', {}).get('drillstring', None)
        casing_id = first_wits.get('metadata', {}).get('casing', None)

        if not drillstring_id and not casing_id:
            return []

        if drillstring_id:
            active_string_type = 'drillstring'
            active_string_id = drillstring_id
        elif casing_id:
            active_string_type = 'casing'
            active_string_id = casing_id

        if previous_string == active_string_id:
            return []

        asset_id = self.global_state['asset_id']
        main_structure = self.build_empty_output(first_wits)

        warning = None
        last_document = None

        if active_string_type in running_string:
            if active_string_type not in reset_config:
                return []
        else:
            message = "This module does not run while {} is the active string in the well.".format(active_string_type)
            warning = {
                'message': message
            }

            if active_string_type not in reset_config and self.collection:
                # get the last record
                worker = API()
                last_document = worker.get(
                    path="/v1/data/corva", collection=self.collection, asset_id=asset_id,
                    sort="{timestamp: -1}", limit=1,
                ).data
                last_document = last_document[0] if last_document else None
        res = self.configure_output_at_config_change(
            main_structure, active_string_type, active_string_id, warning, last_document
        )

        return [res]

    def configure_output_at_config_change(
        self, main_structure, active_string_type, active_string_id, warning=None, last_document=None
    ):
        extra_elements = self.create_output_for_new_config(active_string_type, active_string_id)
        main_structure['data'] = operations.merge_dicts(main_structure['data'], extra_elements)

        if warning:
            main_structure['data'] = operations.merge_dicts(main_structure['data'], {'warning': warning})

        if last_document:
            main_structure['data'] = operations.merge_dicts(last_document.get('data', {}), main_structure['data'])

        return main_structure

    def create_output_for_new_config(self, active_string_type, active_string_id):
        self.set_state('active_string_type', active_string_type)
        self.set_state('active_string_id', active_string_id)

        drillstring_number = None
        if active_string_type == 'drillstring':
            drillstring = get_drillstring_by_id(active_string_id)

            if drillstring:
                drillstring_number = drillstring.get('data', {}).get('id', None)

        self.set_state('drillstring_number', drillstring_number)

        return self.get_config_properties()

    def get_config_properties(self):
        properties = {
            'active_string_type': self.state.get('active_string_type', ''),
            'active_string_id': self.state.get('active_string_id', ''),
            'drillstring_number': self.state.get('drillstring_number', None)
        }

        return properties



# flake8: noqa: E402

"""
This test is created to make sure the framework for local testing
is working correctly.
"""

# added the main directory as the first path
from os.path import dirname
import sys
parent = dirname(dirname(__file__))
sys.path.insert(0, parent)

from worker.test import AppTestRun
from testcase.app import app_lambda


if __name__ == '__main__':
    collections = ['drilling-efficiency.mse']
    # loading of constants should happen here to avoid conflicts
    from testcase.app.app_constants import constants
    app = AppTestRun(app_lambda.lambda_handler, collections, constants)
    app.run()



from setuptools import find_packages, setup


classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Topic :: Software Development :: Libraries',
]

setup(
    name='corva-worker-python',
    author='Jordan Ambra <jordan.ambra@corva.ai>, Mohammadreza Kamyab <m.kamyab@corva.ai>',
    url='https://github.com/corva-ai/corva-worker-python',
    version='0.7.6',
    classifiers=classifiers,
    description='SDK for interacting with Corva',
    keywords='corva, worker',
    packages=find_packages(exclude=["test"]),
    install_requires=["numpy", "redis", "requests", "cached-property"],
    include_package_data=True,
    license='The Unlicense',
)



"""
This file is used to trigger lambda functions locally and test
the results of the app on actual data to make sure of the results.
An example to use this functionality is like the one shows on
'app_test.py' and then that file can run in command line.

Note: make sure you run your tests on QA environment.

== app_test.py file
# added the main directory as the first path
from os.path import dirname
import sys
parent = dirname(dirname(__file__))
sys.path.insert(0, parent)

if __name__ == '__main__':
    collections = ['collection_to_delete']
    app = AppTestRun(lambda_function.lambda_handler, collections)
    app.run()

> python app_test -a 16886 -d True
"""

import argparse
import os
from distutils.util import strtobool
from typing import List

from dotenv import load_dotenv
from tqdm import tqdm

from worker.data.operations import delete_collection_data_of_asset_id, get_one_data_record, gather_data_for_period, \
    point_main_envs
from worker.state.mixins import StateMixin
from worker.test.utils import get_last_processed_timestamp, create_scheduler_events


load_dotenv(override=True)


def generate_runner_parser():
    """
    Creating the supporting arguments
    :return:
    """
    parser = argparse.ArgumentParser(description="Run your tests on an asset.")
    parser.add_argument("-v", "--environment", "--env", type=str, required=False,
                        help="environment, options: 'local', 'qa', 'staging', 'production'")
    parser.add_argument("-a", "--asset_id", "--id", type=int, required=True,
                        help="set asset_id")
    parser.add_argument("-s", "--start_timestamp", "--start", type=int, required=False, default=None,
                        help="start timestamp")
    parser.add_argument("-e", "--end_timestamp", "--end", type=int, required=False, default=None,
                        help="end timestamp")
    parser.add_argument("-i", "--timestep", "--step", type=int, required=False, default=60,
                        help="trigger the lambda function once every step")
    parser.add_argument("-d", "--to_delete", "--delete", type=strtobool, required=False, default=False,
                        help="to delete the state and data")
    parser.add_argument("-g", "--storage_type", "--storage", type=str, required=False, default="redis",
                        help="type of state storage, redis or mongo")
    return parser


class AppTestRun:
    def __init__(self, lambda_handler, collections: List[str], constants=None, args: List[str] = None):
        """
        :param lambda_handler: lambda handler function to run
        :param collections: collections to erase in case to_delete is on
        :param constants: constants module of the whole app, if there is only one app in the whole repository
        pass nothing and the code takes care of that
        :param args: arguments for the run
        """
        self.lambda_handler = lambda_handler
        self.collections = collections

        if not constants:
            from worker import constants

        self.constants = constants

        self.event_type = None
        self.progress = None

        if args is None:
            parser = generate_runner_parser()
            args = parser.parse_args()

        self.initialize(args)

    def initialize(self, args):
        environment = args.environment

        # directing main environment variables to the provided env: API_KEY, API_ROOT_URL, CACHE_URL
        point_main_envs(environment)

        asset_id = args.asset_id

        start_timestamp = args.start_timestamp
        end_timestamp = args.end_timestamp
        step = args.timestep
        self.event_type = self.constants.get('global.event-type')
        to_delete = args.to_delete

        # Setup state storage type environment
        state_storage_type = args.storage_type.lower()
        os.environ["STATE_STORAGE_TYPE"] = state_storage_type

        state_keys = construct_state_keys(self.constants, asset_id)

        if not start_timestamp:
            start_timestamp = get_one_data_record(asset_id, timestamp_sort=+1).get('timestamp')

        if not end_timestamp:
            end_timestamp = get_one_data_record(asset_id, timestamp_sort=-1).get('timestamp')

        if to_delete:
            delete_state_data_of_asset_id(asset_id, state_keys, state_storage_type=state_storage_type)
            delete_collection_data_of_asset_id(asset_id, self.collections)
            print(f"Deleted relevant {state_storage_type} states and collections for this asset!")

        start_timestamp = get_last_processed_timestamp(asset_id, state_keys[0], state_storage_type) or start_timestamp
        print(f"asset_id: {asset_id}, timestamp interval: [{start_timestamp}, {end_timestamp}]")

        events = create_scheduler_events(asset_id, start_timestamp, end_timestamp, step)

        self.progress = tqdm(events, ncols=150)

    def run(self):
        print("\nRunning the main module started ...")
        for event in self.progress:
            schedule_time = str(int(event[0][0]['schedule_start'] / 1000))
            if self.event_type == "wits_stream":
                wits = gather_data_for_period(
                    int(event[0][0]['asset_id']),
                    int(event[0][0]['schedule_start'] / 1000),
                    int(event[0][0]['schedule_end'] / 1000)
                )
                if not wits:
                    continue
                event = [wits]
            self.lambda_handler(event, None)
            self.progress.set_description(schedule_time)

        # Reset environment for other tests
        del os.environ["STATE_STORAGE_TYPE"]


def construct_state_keys(consts, asset_id: int) -> List[str]:
    """
    Get the constants dict of an app and construct the storage keys.
    Note: the first key is the global app key.
    :param consts: app constants module
    :param asset_id: well asset id
    :return: a list of storage keys
    """
    app_key = consts.get("global.app-key")
    module_keys = list(consts.get(app_key).keys())
    state_app = StateMixin()
    global_state_key = state_app.get_formatted_state_key(asset_id, app_key)

    module_state_keys = [
        state_app.get_formatted_state_key(asset_id, app_key, module)
        for module in module_keys
    ]

    return [global_state_key, *module_state_keys]


def delete_state_data_of_asset_id(asset_id: int, state_keys: List[str], state_storage_type=None):
    """
    Delete the state from given storage type
    :param asset_id:
    :param state_keys:
    :param state_storage_type:
    :return:
    """
    state_app = StateMixin(state_storage_type)
    state_app.delete_states(asset_id, state_keys)



class TwoWayDict(dict):
    def __setitem__(self, key, value):
        # Remove any previous connections with these values
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        dict.__setitem__(self, key, value)
        dict.__setitem__(self, value, key)

    def __delitem__(self, key):
        dict.__delitem__(self, self[key])
        dict.__delitem__(self, key)

    def __len__(self):
        """Returns the number of connections"""
        return dict.__len__(self) // 2



from worker.app.modules.activity_module import ActivityModule


class DepthActivityModule(ActivityModule):
    """
    This is an abstract module that needs to be extended by actual modules.
    An example of this module would be running an app every 1 ft.
    """
    module_key = "depth_activity_module"
    collection = "depth_activity_collection"

    # override
    def process_module_state(self, state):
        state = super().process_module_state(state)

        if self.global_state.get("reset", False):
            # todo
            pass

        return state



import worker
import importlib
importlib.reload(worker)
from worker import constants  # noqa: #402


constants.update({
    "global": {
        "app-name": "WorkerTest.StateMixin",
        "app-key": "worker_test-state_mixin",
        "event-type": "wits_stream",
        "query-limit": 3600
    },
    "worker_test-state_mixin": {
        "mongo-tester": {},
        "redis-tester": {},
        "migration-tester": {
            "running_frequency": 0  # just for testing with trigger module
        }
    }
})
